/* BeurlingPack.cpp	By Ken Stephenson, Univ. of Tenn. 2004 */

/* Standalone program; compile with 

   gcc -o BP -I${HOME}/CP/include -L${HOME}/CP/lib -L${HOME}/CP/lib/libCP_libs.a  BeurlingPack.cpp

Related to Beurling's extension of the Riemann Mapping Theorem.
This says that if h(z) is a positive, continuous, bounded function on 
the plane and log(1/h) is subharmonic, then there is a univalent 
function f on the unit disc so that the limit of |f'(z)|-h(f(z)) --> 0 
as z --> e^{it} on unit circle, and f is unique if f(0)=0 and f'(0)>0.

Idea is to read in a 'guide' function g whose image packing is a 
region Omega. Then we read in another packing p, possibly a 
different complex K, which lives in Omega.

Commands: (in shell or from CirclePack)

  g     Read the eucl guide packing, 'guide.p', and its eucl
	max packing 'max_guide.p', store the sharp function.
	Idea is to read q just once.

  p     read the eucl target packing 'target.p', use locations 
	of bdry centers to get value of guide's ratio function, 
	write results to 'ratio_data.q'.

  ?     commands.

  bye   exit

What we are doing is using the guide to, in essense, give us the
function h on Omega (without the log). Note that the ratio function,
sh(g) is defined on the disc, so we are really using sh(g) composed
with g^{-1} when we use the ratio values in the range. Tranferring
these to the target is like setting |f'| on the boundary. Of course, 
after repacking, the boundary circles of p will be in new locations,
so we have to reread p and reset again. The hope is that iteration
will lead to a solution.

One gets this program running in a shell. The packings are saved
from CirclePack, read in here, and commands issued here to save
ratio_data.q. This is read in in 'beurling_flow' (ftn=1). 
Expect to repeat, so keep the guide data. We have gd, which is
the max packing for g converted to euclidean, gr which is the eucl
range packing for g, and we have p, which is the eucl range packing
for our target complex K.

*/

//#include <pwd.h> //FA
//#include <sys/types.h> //FA
//#include <unistd.h> //FA
#include "sa_types.h"
#include "sa_proto.h"

int get_next_cmd(),msg(),emsg();

/* guidedomain, guiderange, p domain, p range */
struct p_data *gd,*gr,*p,*PackData;  

char emsgbuf[BUFSIZE+10],*msgbuf;
   /* Note: msgbuf, emsgbuf, and msg(), emsg() defined here
      are necessary for library routines */

char nextcmd[BUFSIZE],nxt[BUFSIZE];
	/* utility buffers for strings. */

FILE *fpg,*radfp,*rfp;

int main(int argc,char *argv[])
{
  int j,v,vert,pnum,face,hits,flag=0,outofrange=0;
  char *endptr,*cmdptr;
  double *sharp=NULL,sm;
  complex z;
  struct Vertlist *bdrylist=NULL,*vtrace,*hitlist=NULL;

  fpg=radfp=NULL;
  msgbuf=emsgbuf+10; 
  strcpy(emsgbuf,"**ERROR** "); 
		
  /* ---------- initialization --------------- */

  PackData=(struct p_data *)
    calloc(3,sizeof(struct p_data));
  alloc_pack_space(&PackData[0],500,0);
  gd=&PackData[0]; /* guide domain packing */
  gr=&PackData[1]; /* guide range packing */
  p=&PackData[2]; /* target packing range */
  
  /* ---------- hello world ------------------ */

  sprintf(msgbuf,"Remote Program BeurlingPack: set bdry radii of "
	  "'target.p' based on ratio function of 'guide.p'");
  msg();
  
  /* --------- cycle through commands --------- */
  
  while (get_next_cmd()) { /* from user */
    cmdptr=nextcmd;
    if (grab_next(&cmdptr,nxt)) {

/* g read and process guide packing */

      if (!strncmp(nxt,"g",1)) {
	/* read guide pack and its max pack; both should be eucl. */
	flag=0;
	if (!(fpg=fopen("guide.p","r")) || !readpack(fpg,gr)
	    || gr->hes) {
	  sprintf(msgbuf,"BP: read of 'guide.p' failed or pack is not eucl.");
	  emsg();
	}
	else {
	  flag=1;
	  gr->status=1;
	  complex_count(gr,0);
	  facedraworder(gr,0);
	  for (j=1;j<=gr->facecount;j++) gr->faces[j].plot_flag=1;
}
	if (fpg) fclose(fpg);
	if (flag &&!(fpg=fopen("max_guide.p","r")) || !readpack(fpg,gd)
	    || gd->hes || gd->nodecount!=gr->nodecount) {
	  sprintf(msgbuf,"BP: read 'max_guide.p' failed, pack is not eucl, "
		  "or size is different from quide.p.");
	  emsg();
	}
	else {flag=2;gd->status=1;}
	if (flag>1) {
	  sprintf(msgbuf,"BP: read guide.p and max_guide.p; nodecount=%d",
		  gr->nodecount);
	  msg();
	  if (sharp) free(sharp);
	  sharp=(double *)calloc((size_t)(gd->nodecount+1),sizeof(double));
	  for (j=1;j<=gd->nodecount;j++) 
	    sharp[j]=gr->packR_ptr[j].rad/gd->packR_ptr[j].rad;
	}
      }
    
/* p read target packing and write out ratio_data. */    

      else if (strncmp(nxt,"p",1)==0) {
	if (!sharp) { /* no guide function is in place */
	  sprintf(msgbuf,"BP: must first load the 'guide' function.");
	  emsg();
	}
	else {
	  flag=0;
	  if ((fpg=fopen("target.p","r"))==NULL || !readpack(fpg,p)) {
	    sprintf(msgbuf,"BP: read of 'target.p' failed.");
	    emsg();
	  }
	  else {flag=1;p->status=1;}
	  if (fpg) fclose(fpg);
	}
	if (flag) {
	  facedraworder(p,0);
	  sprintf(msgbuf,"BP: read of target packing has "
		  "succeeded; nodecount=%d",p->nodecount);
	  msg();
	  if (p->hes) geom_to_e_pack_convert(p);
	  vert_free(&bdrylist);
	  bdrylist=Node_link_parse(p,"b",&endptr,&hits,
				   &Vlist,&Elist,&Flist,
				   &region,pathlist,pathlength);
	  
	/* open file for ratio data output */
	  if ((rfp=fopen("ratio_data.q","w"))) {
	    fprintf(rfp,"CHECKCOUNT: %d\nRATIOS:\n",p->nodecount);
	    outofrange=0;
	    vtrace=bdrylist;
	    while(vtrace) {
	      z=p->packR_ptr[(v=vtrace->v)].center;
	      /* check if it lies in some circle first */
	      if ((hitlist=cir_search(gr,z))) {
		vert=hitlist->v;
		fprintf(rfp,"%d  %lf\n",v,sharp[vert]);
		vert_free(&hitlist);
	      }
	      else {
		hitlist=tri_search(gr,z);
		if (hitlist) {
		  face=hitlist->v;
		  sm=0.0;
		  for (j=0;j<3;j++) 
		    sm += sharp[gr->faces[face].vert[j]];
		  fprintf(rfp,"%d  %lf\n",v,sm/3.0);
		  vert_free(&hitlist);
		}
		else outofrange++; /* no sharp value */
	      }
	      vtrace=vtrace->next;
	    } /* end of while */
	    fprintf(rfp,"(done)\nEND\n");
	    fclose(rfp);
	    sprintf(msgbuf,"BP: Ratio values have been stored in "
		    "'ratio_data.q'.");
	    msg();
	    if (outofrange) {
	      sprintf(msgbuf,"BP: %d target centers were out of range.",
		      outofrange);
	      emsg();
	    }
	  }
	  else {
	    sprintf(msgbuf,"BP: failed to open 'ratio_data.q' "
		    "for the output.");
	    emsg();
	  }
	}
	else {
	  sprintf(msgbuf,"BP: failed to read 'target.p'");
	  emsg();
	}
      }  
    
/* quit */
      
      else if (!strncmp(nxt,"quit",4) || !strncmp(nxt,"bye",3))
	exit(0);
      
      /* ? help */
      
      else if(strncmp(nxt,"?",1)) {
	sprintf(msgbuf,"guide: read eucl files 'guide.p' and 'max_guide.p'.");
	msg();
	sprintf(msgbuf,"packing: read eucl packing 'target.p'. "
		"New boundary radii will be in 'ratio_data.q'.");
	msg();
	sprintf(msgbuf,"bye");
	msg();
      }
    }
  }  /* end of while */
} /* end of main */

int get_next_cmd()
/* fill nextcmd from parent (or user) */
{
	nextcmd[0]='\0';
	printf("Enter next command (or ?):");
	if (fgets(nextcmd,BUFSIZE,stdin))
		  return 1;
	else return 0;
} /* get_next_cmd */

/* have to provide buffers emsgbuf (and msgbuf, part of it) and
the routines emsg() and msg() because library routines will need
them */

int msg()
{
  fprintf(stdout,"%s\n",msgbuf);
  return 1;
} /* msg */

int emsg()
{
  fprintf(stdout,"%s\n",emsgbuf);
  return 1;
} /* emsg */

int repack_activity_msg(char *msg)
{
  printf("*");
  return 1; /* dummy call */
}
