#include "cp_types.h"
#include "cp_graphics_proto.h"
#include "cp_proto.h"
#include "cp_head.h"

/* Run with

   Layout_lights -p {parentpacking}.p -q {radii}.q {lightpackings}.vl

Last argument is a *.vl file of (possibly: parent packing) and 
light subpackings; optional

* designate parent packing (overridden if parent is in *.vl file)
* file of radii, which overrides those in parent packing.

Goal: use radii to compute layouts for each of the light packings,
store centers of those layouts in an output file *.vl.ctr. 
(May want to append these to the *.vl file.)

Compile with circle packing libraries.
*/

char *emsgbuf,*msgbuf;
Mobius Mob;

int main(int argc,char *argv[])
{
  int i,j,N=0,overlap,v,w,count=0;
  struct p_data *p,*p_tmp;
  struct K_data *pK_ptr,*ptK_ptr;
  struct R_data *pR_ptr,*ptR_ptr;
  struct Edgelist *vtrace=NULL;
  struct p_light **plight;
  char line[BUFSIZE],filename[NAME_MAX],tmpname[NAME_MAX];
  char parent[NAME_MAX],radfile[NAME_MAX];
  FILE *fp=NULL,*fp_out=NULL;
  char center_re_str[MAX_ACCUR + 25];
  char center_im_str[MAX_ACCUR + 25];

  set_CPack_globals(); /* set up library global variables */

  /* ----------- user code starts here */

  p=init_packing();
  p_tmp=init_packing();
  printf("Layout_lights: for reading a *.vl file, parent file,\n"
	 "and radius file; computing light packing layouts, \n"
	 "recording the packings in separate files..\n");
  *filename=*radfile=*parent='\0';
  for (i=1;i<argc;i++){
    if (!strcmp(argv[i],"-p") && i<(argc+1)) 
      strcpy(parent,argv[++i]);
    else if (!strcmp(argv[i],"-q") && i<(argc+1)) 
      strcpy(radfile,argv[++i]);
    else if (!strcmp(argv[i],"-v") && i<(argc+1)) 
      strcpy(filename,argv[++i]);
  }
  if (!strlen(filename)){ /* no filename yet? default is the last argv */
    if (argc>1)
      strcpy(filename,argv[argc-1]);
    if (!strlen(filename)){ /* didn't get a filename */
      printf("Enter a *.vl file (with parent and radii):\n");
      fgets(line,sizeof(line),stdin);
      if (!sscanf(line,"%s",filename)){ /* still no filename! */
	printf("Failed to get a *.vl filename. Exiting.\n");
	exit(1);
      }
    }
  }
  if (!(N=read_parent_child(filename,p,&plight, &overlap))){
    printf("Failed to open %s.",filename);
    exit(2);
  }
  if (strlen(parent)){ /* read parent (override parent from *.vl) */
    if (!(fp=fopen(filename,"r"))
	|| !(readpack(fp,p))){
      printf("Error: couldn't read given parent %s.\n",parent);
      if (!p->nodecount){
	printf(" and I got no parent from %s, so I'll quit.\n",filename);
	exit(3);
      }
    }
    if (fp) {fclose(fp);fp=NULL;}
  }
  if (strlen(radfile)){ /* read radii from specified file */
    if (!(fp=fopen(filename,"r"))
	|| !(readpack(fp,p))){
      printf("Warning: did not get radii from %s.\n",radfile);
    }
    if (fp) {fclose(fp);fp=NULL;}
  }
  pK_ptr=p->packK_ptr;
  pR_ptr=p->packR_ptr;

  for (i=1;i<=N;i++)
    {
      alloc_pack_space(p_tmp,5000,0);
      if (convert_from_p_light(p_tmp,plight[i],0))
	{
	  ptK_ptr=p_tmp->packK_ptr;
	  ptR_ptr=p_tmp->packR_ptr;
	  /* choose alpha vertex deep inside, set gamma, too */
	  for (j=1;j<=p_tmp->nodecount;j++){
	    if (ptK_ptr[j].bdry_flag) ptK_ptr[j].util_flag=1;
	    else ptK_ptr[j].util_flag=0;
	  }
	  p_tmp->alpha=1;
	  int *list=label_generations(p_tmp,-1,&p_tmp->alpha,&count);
	  free(list);
	  if (!count || !p_tmp->alpha){
	    printf("Error in counting generations to find alpha.\n");
	    exit(2);
	  }
	  choose_gamma(p_tmp);
	  /* set default radii, but then transfer radii from parent */
	  for (j=1;j<=p_tmp->nodecount;j++) ptR_ptr[j].rad=.5;
	  vtrace=p_tmp->vertex_map;
	  count=0;
	  while (vtrace){
	    if ((v=vtrace->v)>0 && v<=p_tmp->nodecount 
		&& (w=vtrace->w)>0 && w<=p->nodecount) {
	      ptR_ptr[v].rad=pR_ptr[w].rad;
	      count++;
	    }
	    vtrace=vtrace->next;
	  }
	  if (count!=p_tmp->nodecount)
	    printf("Only got %d out of %d radii in subpack %d.",
		   count,p_tmp->nodecount,N);
	  /* compute the layout using current radii */
	  if (!fix_call(p_tmp,"")); 
	  /* record the packing */
	  sprintf(tmpname,"%s.%d.p",filename,i); /* file name extension */
	  if (!(fp_out=fopen(tmpname,"w"))
	      || !writepack(fp_out,p_tmp,00133,0))
	    {
	      printf("Error opening or writing output file %s.",tmpname);
	      exit(1);
	    }
	  fclose(fp_out);
	  printf("Just saved file %s.",tmpname);
	} /* loop through light packings */
    }

  /* ----------- user code ends here */
  exit(0);
} /* main */


/* ===== required communication routines; user may want to modify === */

int emsg()
{
  printf(emsgbuf);return 1;
} /* emsg */

int msg()
{
  printf(msgbuf);return 1;
} /* msg */

int repack_activity_msg(char *datastr)
     /* report progress of repacking */
{
  strcpy(msgbuf,datastr);
  msg();
  return 1;
} /* repack_activity_msg */

int refresh_canvas(struct s_data *q)
     /* some drawing routines want to refresh */
{return 1;} /* refresh_canvas */

/* ================================================================= */

