#include "sa_types.h"
#include "sa_proto.h"

/* add layer of nodes to bdry segment of pack p from one vertex v1 to v2.
 * Can specify making vertices degree n (default) or merely 
 * adding vert-for-vert (flag -t). Return number of new vertices added. 
 * In default case, flag returns 0 if at least one node ends up with more 
 * than n neighbors. 
 * 
 * Afterwards, none of original verts should be on boundary; no new edges
 * should be added between existing boundary verts.
 * 
 * datastr: v1 v2 n or datastr: v1 v2 -t. 
 */

#define SEPARATOR 1 /* separator symbol is ';' */
#define PACK 2
#define FLAG 3
#define STRING 4
#define INTEGER 5
#define FLOAT 6

int add_layer(struct p_data *p,char *datastr,int *flag)
{
	int v1,v2,v3,eqflag=0,d,count=0,need,vert,nextvert,i,allocation,N,l;
	char *nextpoint;
	StrTok *strTok=NULL;
	
	if (!p->num_bdry_comp) {
    	sprintf(msgbuf,"Can't apply 'add_layer'; there are no boundary vertices.");
		emsg();
	    return 0;
	}

	if (!(strTok=string_tokenizer(datastr,&allocation,&N)) || N<1) return 0;
	if (N<3 || strTok[0].key!=INTEGER || strTok[1].key!=INTEGER) {
		tok_free(&strTok,allocation);
		sprintf(msgbuf,"Check usage: add_layer v1 v2 n (or -t).");
		emsg();
		return 0;
	}
	v1=strTok[0].intvalue;
	v2=strTok[1].intvalue;
	if (!p->packK_ptr[v1].bdry_flag || !(l=verts_share_bdry(p,v1,v2))) {
		tok_free(&strTok,allocation);
		sprintf(msgbuf,"Check usage: add_layer v1 v2 n (or -t). v1/v2 on bdry?");
		emsg();
		return 0;
	}
	if (strTok[2].key==INTEGER) { // default situation: d is desired degree
		d=strTok[2].intvalue;
		if (d<4 || d>MAX_PETALS) {
			tok_free(&strTok,allocation);
			sprintf(msgbuf,"Check usage: add_layer v1 v2 n. Your n is not legal");
			emsg();
			return 0;
		}
	}
	else if (tok_is_flag(strTok[2],'t')==1) d=0; // -t flag case: set n=0 as flag
	else { // error: no other options currently planned
		tok_free(&strTok,allocation);
		sprintf(msgbuf,"Check usage: add_layer v1 v2 n or add_layer v1 v2 -t");
		emsg();
		return 0;
	}
	tok_free(&strTok,allocation);

	if (v2==v1 || v2==p->packK_ptr[v1].flower[p->packK_ptr[v1].num]) {
    	eqflag=1;
	    v2=p->packK_ptr[v2].flower[p->packK_ptr[v2].num];
	}
	// add to get degree d; every circle, v1 to v2 inclusive,
	//   must get at least one new neighbor; we never just enclose
	//   since that identifies two circles of the original bdry.
	//   (Of course, the new circle may have been added already in
	//   a previous step.)
	*flag=1;
	if (d>0) { 
		vert=v1;
		nextvert=p->packK_ptr[vert].flower[0];
		count += add_vert(p,vert); // new circle shared with upstream nghb.

    	/* go until you get to v2 */
		while (vert!=v2 && p->packK_ptr[vert].bdry_flag) {
	      if (d<(p->packK_ptr[vert].num+1)) *flag=0;
	      else if ((need=d-p->packK_ptr[vert].num-1)>0)
			for (i=1;i<=need;i++)
			  count += add_vert(p,vert);
	      enfold(p,vert);
    	  vert=nextvert;
	      nextvert=p->packK_ptr[vert].flower[0];
	      if (vert==v1) { // shouldn't happen
	      	sprintf(msgbuf,"error in 'add_layer' processing; exit");
	      	emsg();
	      	return 0;
	      }
		}
		if (vert==v2) { /* now do v2 itself */
    		if (d<(p->packK_ptr[vert].num+1)) *flag=0;
	      	else if ((need=d-p->packK_ptr[vert].num-1)>0)
				for (i=1;i<=need;i++)
				  count += add_vert(p,vert);
			enfold(p,vert);
		}
	}
	else { // this is -t flag situation
		vert=v1;
		nextvert=p->packK_ptr[vert].flower[0];
		count += add_vert(p,vert);
		while (vert!=v2 && p->packK_ptr[vert].bdry_flag) {
			count+=add_vert(p,vert);
			enfold(p,vert);
			vert=nextvert;
			nextvert=p->packK_ptr[vert].flower[0];
			if (vert==v1) { // shouldn't happen
	      		sprintf(msgbuf,"error in 'add_layer' processing; exit");
		      	emsg();
		      	return 0;
			}
		}
		if (vert==v2) {
			if (eqflag) 
				enfold(p,vert);
			else {
				count += add_vert(p,vert);
				enfold(p,vert);
			}
		}
	}
		
	complex_count(p,0);
	facedraworder(p,0);
	set_aim_default(p);
	return count;
} /* add_layer */

/**
 * For parsing requests to add multiple full generations of bdry verts.
 * Return 0 on error.
 */
int add_generation(struct p_data *p,char *datastr)
{
	int d,i,j,n,hits,N,allocation,count=0,tflag=0;
	StrTok *strTok=NULL;

    if (!p->num_bdry_comp) {
       sprintf(msgbuf,"Can't apply 'add_gen'; packing has no boundary vertices.");
       emsg();
       return 0;
    }

	if (!(strTok=string_tokenizer(datastr,&allocation,&N)) || N<1) return 0;
	for (i=1;i<N;i++) if (tok_is_flag(strTok[i],'t')==1) tflag=1;
	
	if (tflag) { // circle-for-circle mode
		if (strTok[0].key!=INTEGER) {
			tok_free(&strTok,allocation);
			sprintf(msgbuf,"Check usage: add_gen n -t -b {w..}.");
			emsg();
			return 0;
		}
		n=strTok[0].intvalue;
	}
	else { // standard: get number of generations to add and degree.		
		if (N<2 || strTok[0].key!=INTEGER || strTok[1].key!=INTEGER) {
			tok_free(&strTok,allocation);
			sprintf(msgbuf,"Check usage: add_gen n d -b {w..}.");
			emsg();
			return 0;
		}
		n=strTok[0].intvalue;
		d=strTok[1].intvalue;
		if (n<1 || d<3 || d>MAX_PETALS) {
    		sprintf(msgbuf,"Check usage: add_gen n d -b {w..}.");
			emsg();
		    tok_free(&strTok,allocation);
			return 0;
		}
	}
	
	// start with full boundary list; cut down later if -b flag is set
	struct Vertlist *blist=NULL,*btrace;
	blist=btrace=(struct Vertlist*)calloc(1,sizeof(struct Vertlist));
	for (j=1;j<=p->num_bdry_comp;j++) {
	   	btrace=btrace->next=(struct Vertlist*)calloc(1,sizeof(struct Vertlist));
	   	btrace->v=p->bdry_starts[j];
	}
    btrace=blist;
    blist=blist->next;
    free(btrace);

	// look for -b flag to indicate boundary 'start' vertices
	//   in case of multiple boundary components: default is 'all'
	for (i=1;i<N;i++) {
	    if (tok_is_flag(strTok[i],'b')==1) { // got it; now vertices
		   	char *bstr=flag_to_flag(strTok,i+1,N);
		   	char *endptr;
    		blist=Node_link_parse(p,bstr,&endptr,&hits,
			  &Vlist,&Elist,&Flist,&region,pathlist,pathlength);
			free(bstr);
	    }
	}
   
	if (blist==NULL) { // didn't get any components?
		sprintf(msgbuf,"Check 'add_gen' call; no boundary components were given");
	    emsg();
	    tok_free(&strTok,allocation);
	    return 0;
	}
	
    // repeat n layers
    int jj=1,localhits;
    for (i=1;i<=n;i++) {
	    btrace=blist;
		while (btrace) {
			localhits=1;
			if (p->packK_ptr[btrace->v].bdry_flag) {
				if (tflag) sprintf(next,"%d %d -t",btrace->v,btrace->v);
				else sprintf(next,"%d %d %d",btrace->v,btrace->v,d);
				localhits=add_layer(p,next,&hits);
		    	jj *= hits; // 0 ==> some degrees exceed d
				if (localhits>0) {
					count +=localhits;
					btrace->v=p->nodecount; // must be vert on this new bdry
				}
			}
			btrace=btrace->next;
			if (localhits==0) { // something went wrong, break off
				i=n+1;
				btrace=NULL;
			}
		}
	}
		
	// get any degrees exceeding d??
    if (!tflag && !jj) {
		   sprintf(msgbuf,"add_gen: Some circles exceeded %d degree.",d);
		   emsg();
		}
	tok_free(&strTok,allocation);
	vert_free(&blist);
	return count;		
} /* add_generation */

