/*
 * bio.c:  Provides a subset of stdio library calls that
 * call either read/write or send/recv to perform underlying
 * I/O operations, depending on whether the underlying 'handle'
 * is a socket or a file descriptor.  This is needed since using
 * read/write on sockets doesn't work in WIN32.  Note that the
 * bsopen call, which generates an IOBUF from a socket, can only
 * be called on a connection-oriented socket (i.e. TCP only).
 *
 * Currently, these routines assume no one will ever want to
 * print anything longer than 8kbyte in a single call (blech).
 * New functions will be added as needed, but it is unlikely
 * that the entire stdio library API will be supported.  In
 * particular, it appears impossible to implement a portable
 * scanf wrapper, meaning that bscanf would have to contain
 * most of the code from the actual stdio library implementation.
 */

#include "sa_types.h"
#include "sa_proto.h"
#include "sa_head.h"

#define IOBUF_MAX_CAPACITY 8192

typedef enum buf_t { IOBUF_FD, IOBUF_SOCK } buf_t;

typedef struct TRUE_IOBUF  {
    int d;           /* A socket or a file descriptor */
    SOCKET s;
    buf_t bufType;   /* Determines whether we should perform 
			IO using read/write or send/recv */
    char *buf;
    int nchars;      /* Number of characters already in the buffer */
    int start;       /* Beginning of valid data in the buffer */
}  TRUE_IOBUF;

int getcharIfNeeded(TRUE_IOBUF *iobuf, int index);

void brealign(TRUE_IOBUF *iobuf)
{
    memmove(iobuf->buf, iobuf->buf + iobuf->start, iobuf->nchars);
    iobuf->start = 0;
} /* brealign */

IOBUF *bdopen(int d, const char *mode)
{
    TRUE_IOBUF *newiobuf = (TRUE_IOBUF *) malloc(sizeof(TRUE_IOBUF));

    if(newiobuf == NULL)  return NULL;

    newiobuf->d = d;
    newiobuf->bufType = IOBUF_FD;
    newiobuf->buf = (char *) malloc(sizeof(char) * IOBUF_MAX_CAPACITY);
    if(newiobuf->buf == NULL)  {
        free(newiobuf);
        return NULL;
    }
    newiobuf->nchars = 0;
    newiobuf->start = 0;

    return (IOBUF *) newiobuf;
} /* bdopen */

IOBUF *bsopen(int s, const char *mode)
{
    TRUE_IOBUF *newiobuf;

    newiobuf = (TRUE_IOBUF *) malloc(sizeof(TRUE_IOBUF));
    if(newiobuf == NULL)  return NULL;

    newiobuf->s = s;
    newiobuf->bufType = IOBUF_SOCK;
    newiobuf->buf = (char *) malloc(sizeof(char) * IOBUF_MAX_CAPACITY);
    if(newiobuf->buf == NULL)  {
        free(newiobuf);
        return NULL;
    }
    newiobuf->nchars = 0;
    newiobuf->start = 0;

    return (IOBUF *) newiobuf;
} /* bsopen */

int bclose(IOBUF *stream)
{
    TRUE_IOBUF *oldiobuf = (TRUE_IOBUF *) stream;

    bflush(stream);
    if(oldiobuf->bufType == IOBUF_FD)  {
#ifdef WIN32
        _close(oldiobuf->d);
#else
        close(oldiobuf->d);
#endif
    }
    else  {
        closesocket(oldiobuf->s);
    }
    return 0;
} /* bclose */

int bprintf(IOBUF *stream, const char *format, ...)
{
	//return 0; // FA
	//strcat(MSG_OUT_FP,format);
	//return 0;
	
    TRUE_IOBUF *iobuf = (TRUE_IOBUF *) stream;
    int retval;
    va_list ap;

	bflush(stream);

    va_start(ap, format);
    retval = vsprintf(iobuf->buf + iobuf->start + iobuf->nchars, format, ap);
    iobuf->nchars += retval;
    va_end(ap);

    bflush(stream);

    return retval;
} /* bprintf */

int bflush(IOBUF *stream)
{
	//return 0; // FA!
    TRUE_IOBUF *iobuf = (TRUE_IOBUF *) stream;
    int prefix, len;

    prefix = 0;
    while(prefix < iobuf->nchars)  {
        //if(iobuf->bufType == IOBUF_FD)  
	  len = write(iobuf->d, iobuf->buf + iobuf->start + prefix, 
		      iobuf->nchars - prefix);
        //else  len = send(iobuf->s, iobuf->buf + iobuf->start + prefix, 
		//	 iobuf->nchars - prefix, 0);
        if(len < 0)  return EOF;
        prefix += len;
    }
    iobuf->nchars = 0;
    iobuf->start = 0;

    return 0;
} /* bflush */

char *bgets(char *s, int size, IOBUF *stream)
{
    TRUE_IOBUF *iobuf = (TRUE_IOBUF *) stream;
    int end;

    if(size == 0)  return NULL;
    brealign(iobuf);

    for(end = 0; end < size - 1 && iobuf->start + end < IOBUF_MAX_CAPACITY; 
	end++)  {
        if(getcharIfNeeded(iobuf, end) < 0)  break;
        if(iobuf->buf[end] == '\n')  {
            end++;
            break;
        }
    }

    if(end == 0 && size > 1)  return NULL;

    memcpy(s, iobuf->buf + iobuf->start, end);
    s[end] = '\0';
    iobuf->start += end;
    iobuf->nchars -= end;

    return s;
} /* bgets */

int getcharIfNeeded(TRUE_IOBUF *iobuf, int index)
{
    if(index >= iobuf->nchars)  {
        if(iobuf->bufType == IOBUF_FD)  {
            if(read(iobuf->d, iobuf->buf + iobuf->start + index, 1) <= 0)  
	      return -1;
        }
        else  {
            if(recv(iobuf->s, iobuf->buf + iobuf->start + index, 1, 0) <= 0)
	      return -1;
        }
        iobuf->nchars++;
    }
    return 0;
} /* getcharIfNeeded */
