#include "sa_types.h"
#include "sa_proto.h"

/* Routine to build a connected edge-path which encircles a
specified set of vertices. The archetype use is to define an
edge-path approximating a simple closed curve given in pathlist.
Eventually there should be other uses.

Approach: mark verts on the two sides of the target region,
find transverse edges -- one end in, one out --- and use that
to build up the desired edge-path. 

If init_v is valid vert index, try to start the edge-path there;
if that's not appropriate, try to start at one of its nghbs.

Options (for now):

opt=1: use list of "in" vertices (vertlist)

Return NULL on error; pass back number of edges in "hits".

No guarantee: There can be several strange outcomes: non-simple,
edges transited each way, etc. 

*/

/* these variables used by subroutines at end */
int *blob=NULL;
struct K_data *pK_ptr;
struct R_data *pR_ptr;

struct Edgelist *encircle(struct p_data *p,struct Vertlist *vertlist,
			  int init_v, int *hits,int opt)
{
  int j,v,k,indx,old_v,last_v,next_v,next_one,count=0;
  char *endptr;
  struct Vertlist *trace;
  struct Edgelist *E_list=NULL,*etrace;

  if (!p->status || !vertlist) return NULL;
  blob=NULL; /* wipe out anything left over from earlier run */
  pK_ptr=p->packK_ptr;pR_ptr=p->packR_ptr;
  
  /* mark the blob verts */
  blob=(int *)calloc(p->nodecount+1,sizeof(int));
  trace=vertlist;
  while (trace && count<p->nodecount) {
    blob[trace->v]=1;
    trace=trace->next;
    count++;
  }
  /* fill in any isolated non-blobs */
  for (v=1;v<=p->nodecount;v++) 
    if (!blob[v] && nghb_blob_count(v)==pK_ptr[v].num+pK_ptr[v].bdry_flag) {
      blob[v]=1;
      count++;
    }

  if (!count || count>=p->nodecount) {
    free(blob);
    sprintf(msgbuf,"encircle: given vertlist had all vertices in it.");
    emsg();
    if (p->num_bdry_comp!=1 
	|| !(E_list=Node_pair_link(p,"b",&endptr,hits,&Vlist,&Elist,&Flist,
		   &region,pathlist,pathlength))) {
      sprintf(msgbuf,"encircle: no qualified edge-path");
      emsg();
      return NULL;
    }
    sprintf(msgbuf,"encircle: whole packing enclosed; returned the "
	    "boundary edge-list.");
    msg();
    return E_list;
  }
  
  /* check status of init_v (and nbhbs) to see if we should start there;
     come out with qualified init_v or init_v=0, in which case we 
     proceed as though we had no suggested starting vertex.*/

  *hits=0;
  if (init_v>0 && init_v<=p->nodecount) {
    old_v=init_v;
    init_v=0;
    indx=-1;

    if ((indx=triple_blob(old_v))>=0) init_v=old_v;
    else {
      for (j=0;(j<(pK_ptr[old_v].num+pK_ptr[old_v].bdry_flag) && !init_v);j++)
	if (blob[(k=pK_ptr[old_v].flower[j])] &&
	    (indx=triple_blob(k))>=0) 
	  init_v=pK_ptr[v].flower[j];
    }
	
  }  
  /* If init_v set, then it's 'in' vertex with at least one 'out'
     nghb. If not, let's find the first such case. */

  if (!init_v) {
    for (v=1;(v<=p->nodecount && !init_v);v++) {
      if ((indx=triple_blob(v))>=0) init_v=v;
    }

  if (!init_v || indx<0) { /* some error */
      sprintf(msgbuf,"encircle: no vertices were found on edge of given "
	      "set.");
      emsg();
      free(blob);
      return NULL;
    }
  }

  E_list=etrace=(struct Edgelist *)calloc(1,sizeof(struct Edgelist));

  /* Start edge list with init_v; indx shows direction to start;
     work around the edge of the blob. */

  etrace=etrace->next=(struct Edgelist *)calloc(1,sizeof(struct Edgelist));
  etrace->v=last_v=init_v;
  etrace->w=next_v=pK_ptr[init_v].flower[indx];
  (*hits)++;
  while (next_v!=init_v 
	 && (indx=reverse_blob(next_v,nghb(p,next_v,last_v)))>=0) {
    etrace=
      etrace->next=(struct Edgelist *)calloc(1,sizeof(struct Edgelist));
    etrace->v=last_v=next_v;
    etrace->w=next_v=pK_ptr[last_v].flower[indx];
    (*hits)++;
  }
  if (next_v==init_v) { /* edge-path has returned to its start */
    etrace=E_list;
    E_list=E_list->next; /* first one was empty */
    free(etrace);
    free(blob);
    return E_list;
  }
  else {
    sprintf(msgbuf,"encircle: some processing error, abort.");
    emsg();
    free(blob);
    edge_free(&E_list);
    return NULL;
  }
} /* encircle */

int nghb_blob_count(int v)
  /* count of nghbs marked in blob */
{
  int j,click=0,num;

  num=pK_ptr[v].num+pK_ptr[v].bdry_flag;
  for (j=0;j<num;j++)
    if (blob[pK_ptr[v].flower[j]]) click++;
  return click;
} /* nghb_blob_count */

int triple_blob(int v)
  /* return first index with: petal, next petal in blob, previous
     petal not in blob. -1 on failure. */
{
  int j,num=pK_ptr[v].num;

  if (!blob[v]) return -1;
  if (pK_ptr[v].bdry_flag) {
    for (j=1;j<pK_ptr[v].num;j++)
      if (blob[pK_ptr[v].flower[j]] && blob[pK_ptr[v].flower[(j+1)%num]]
	  && !blob[pK_ptr[v].flower[(j-1+num)%num]])
	return j;
  }
  else {
    for (j=0;j<num;j++)
      if (blob[pK_ptr[v].flower[j]] && blob[pK_ptr[v].flower[(j+1)%num]]
	  && !blob[pK_ptr[v].flower[(j-1+num)%num]])
	return j;
  }
  return -1;
} /* triple_blob */

int reverse_blob(int v,int indx)
  /* came in to v from direciton of ngbh indx (presumably a blob); 
     move clockwise as long as petals are blobs, return indx of
     last blob. Return -1 if very next one (indx-1) is non-blob
     or if v interior and all petals are blobs. */
{
  int j,k,num=pK_ptr[v].num;

  if (pK_ptr[v].bdry_flag) {
    if (indx==0 || !blob[pK_ptr[v].flower[indx-1]]) return -1;
    for (j=indx;j>0;j--)
      if (!blob[pK_ptr[v].flower[j-1]]) return j;
    return 0;
  }
  if (!blob[pK_ptr[v].flower[(indx-1+num)%num]]) return -1;
  for (j=1;j<num;j++) {
    k=(indx-j+num)%num;
    if (!blob[pK_ptr[v].flower[(k-1+num)%num]]) return(k);
  }
  return -1;
} /* reverse_blob */

struct Edgelist *encircle_call(struct p_data *p,char *datastr)
  /* fixup: want more options eventually, like choosing a preferred
     starting vertex for the edgepath. */
{
  int hits;
  char *endptr,*nextpoint=datastr;
  struct Vertlist *vertlist=NULL;
  struct Edgelist *edgelist=NULL;

  stripsp(nextpoint);
  if (!p->status || !(vertlist=Node_link_parse(p,datastr,&endptr,&hits,
                    &Vlist,&Elist,&Flist,&region,pathlist,pathlength))
      || hits>p->nodecount-1) { 
    sprintf(msgbuf,"encircle: usage: give vertice {v..} to be encircled.");
    emsg();
    return NULL;
  }
  edgelist=encircle(p,vertlist,0,&hits,0);
  if (edgelist) {
    sprintf(msgbuf,"encircle: returned edgelist of length %d, initial "
	    "vertex is %d.",hits,edgelist->v);
    msg();
    return edgelist;
  }
  return NULL; 
} /* encircle_call */
