#include "sa_types.h"
#include "sa_proto.h"

/* Generate, put in order, and return list of interior edges 
having selected properties based on cross ratios (x-ratios) of 
ends and common neighbors; default criterion is "illegal" by 
Delaunay criterion --- circle defined by the vertices of
the face on one side contain the face on the other. Can also sort
by distance of x-ratio from that of four in regular hex pattern.
Only works in eucl case at this time.  Returns list which can be 
ordered by various criteria.

  crit_code: (which circles to look at)
     1   Delauney: imaginary part of x-ratio < 0. (default) 
     2   look at all

  order_code: (specified order to list?)
     1   Most negative to most positive imaginary part
     2   largest ratio of radii at ends of the edge
     3   furthest in absolute value from hex ideal, -1/2 + sqrt(3)/2 i

*/

static int order_code,crit_code;
static double hex_thresh;
typedef struct {
  int v1,v2;
  complex x_ratio;
} X_ratios;



void qsort_xrat(struct p_data *p,X_ratios *xrat,
		int left,int right);
void swap_xrat(X_ratios *xrat,int i,int j);
int ck_for_swap(struct p_data *p,X_ratios *xrat,
		int left,int right);

struct Edgelist *legal_edges(struct p_data *p,char *datastr)
{
  int hits,i,j,k,v1,v2,v3,v4,edgecount=0,count=0;
  double e1,e2,e3,e4,e5;
  complex z;
  char *endptr,*nextpoint;
  struct Edgelist *elist=NULL,*etrace,*outlist=NULL,*otrace;
  struct R_data *pR_ptr=p->packR_ptr;
  struct K_data *pK_ptr=p->packK_ptr;  
  X_ratios *xrat=NULL;

  order_code=0;
  crit_code=1;
  if (p->hes){ /* not euclidean? return */
    sprintf(msgbuf,"legal_edge can be used only with euclidean packings.");
    emsg();
    return outlist;
  }
  nextpoint=datastr;
  stripsp(nextpoint);
  while (*nextpoint=='-') { /* pick off flags */
    if (nextpoint[1]=='f') { /* order of list? */ 
      nextpoint += 2;
      if (!grab_next(&nextpoint,next)) { 
	sprintf(msgbuf,"legal_edge: usage -f %%d, code?");
	emsg();
	return outlist;
      }
      if (next[0]=='i') /* most negative imaginary part (default)  */
	order_code=1;
      else if (next[0]=='r') /* ratio of radii (> 1) of edge endpoint radii */
	order_code=2;
      else if (next[0]=='h') {/* furthest from "hex" (-1/2+sqrt(3)/2 i) */
	order_code=3;
	hex_thresh=0.0;
	singledouble tmpht;
	if (*nextpoint!='-' && grab_next(&nextpoint,next) 
	    && sscanf(next,"%lf",&tmpht)) {
	  /* get threshold for distance from hex? */
	  hex_thresh=(double)tmpht;
	  if (hex_thresh<0.0) hex_thresh=0.0;
	}
      }
    }
    else if (nextpoint[1]=='c') { /* criterion? */
      nextpoint += 2;
      if (!grab_next(&nextpoint,next)) { 
	sprintf(msgbuf,"legal_edge: usage -c %%d, code?");
	emsg();
	return outlist;
      }
      if (next[0]=='i') /* Imaginary part of x_ratio negative */
	crit_code=1;
      else if (next[0]=='a') /* all edges */
	crit_code=2;
    }
    stripsp(nextpoint);
  } /* end of while */
  next[0]='a';next[1]='\0';
  /* check for input list, else use all. */
  if ((elist=Node_pair_link(p,nextpoint,&endptr,&hits,&Vlist,&Elist,
			    &Flist,&region,pathlist,pathlength))
      || (elist=Node_pair_link(p,next,&endptr,&hits,&Vlist,&Elist,
			    &Flist,&region,pathlist,pathlength)) ) { 
    edgecount=hits;
    xrat=(X_ratios *)
      calloc((size_t)(edgecount+1),sizeof(X_ratios));
    etrace=elist;
    while (etrace){
      if ((j=nghb(p,(v1=etrace->v),(v3=etrace->w)))>=0 
	  && (k=nghb(p,v3,v1))>=0 
	  && j!=pK_ptr[v1].num && k!=pK_ptr[v3].num){
	v4=pK_ptr[v1].flower[j+1];
	v2=pK_ptr[v3].flower[k+1];
	if (p->overlap_status){ /* have overlaps to use */
	  e1=pK_ptr[v1].overlaps[nghb(p,v1,v2)];
	  e2=pK_ptr[v3].overlaps[k];
	  e3=pK_ptr[v3].overlaps[nghb(p,v3,v4)];
	  e4=pK_ptr[v1].overlaps[j+1];
	  e5=pK_ptr[v1].overlaps[j];
	}
	else e1=e2=e3=e4=e5=1.0;
	z=quad_cross_ratio(pR_ptr[v1].rad,pR_ptr[v2].rad,
		   pR_ptr[v3].rad,pR_ptr[v4].rad,
		   e1,e2,e3,e4,e5);
	if (criterion(z)){ /* edge meets criterion? */
	  xrat[count].v1=v1;
	  xrat[count].v2=v3;
	  xrat[count].x_ratio=z;
	  (count)++;
	}
      }
      etrace=etrace->next;
    } /* end of while */
    edge_free(&elist);
  }

  /* put in order?  */
  if (order_code && count) 
    qsort_xrat(p,xrat,0,(count)-1);
  /* make outlist */
  outlist=otrace=(struct Edgelist *)
    calloc((size_t)1,sizeof(struct Edgelist));
  for (i=0;i<count;i++) {
    otrace=otrace->next=(struct Edgelist *)
      calloc((size_t)1,sizeof(struct Edgelist));
    otrace->v=xrat[i].v1;
    otrace->w=xrat[i].v2;
  }
  otrace=outlist;
  outlist=outlist->next;
  free(otrace);
  free(xrat);
  /* print message */
  if (outlist && (j=nghb(p,(v1=outlist->v),(v3=outlist->w)))>=0 
      && (k=nghb(p,v3,v1))>=0 
      && j!=pK_ptr[v1].num && k!=pK_ptr[v3].num) {
    v4=pK_ptr[v1].flower[j+1];
    v2=pK_ptr[v3].flower[k+1];
    if (p->overlap_status){ /* have overlaps to use */
      e1=pK_ptr[v1].overlaps[nghb(p,v1,v2)];
      e2=pK_ptr[v3].overlaps[k];
      e3=pK_ptr[v3].overlaps[nghb(p,v3,v4)];
      e4=pK_ptr[v1].overlaps[j+1];
      e5=pK_ptr[v1].overlaps[j];
    }
    else e1=e2=e3=e4=e5=1.0;
    z=quad_cross_ratio(pR_ptr[v1].rad,pR_ptr[v2].rad,
		       pR_ptr[v3].rad,pR_ptr[v4].rad,
		       e1,e2,e3,e4,e5);
  }
  sprintf(msgbuf,"Found %d illegal edges; first edge x-ratio = %lf+i*%lf.",
	  count,(singledouble)z.re,(singledouble)z.im);
  msg();
  return outlist;
} /* legal_edges */

int criterion(complex z)
{
  complex HEX_X={-.5,.8660254040};

  if (crit_code==1) { /* z.im<0 */
    if (z.im<0) return 1;
  }
  else if (order_code==3) { /* distance to hex */
    if (cAbs(csub(z,HEX_X))>= hex_thresh) return 1;
  }
  else if (crit_code==2) { /* any edge */
    return 1;
  }
  return 0;
} /* criterion */
    
void qsort_xrat(struct p_data *p,X_ratios *xrat,
		      int left,int right)
  /* copied qsort from K&R. */
{
  int i,last;

  if (left>=right) return;
  swap_xrat(xrat,left,(left+right)/2);
  last=left;
  for (i=left+1;i<=right;i++) {
    if (ck_for_swap(p,xrat,i,left))
      swap_xrat(xrat,++last,i);
  }
  swap_xrat(xrat,left,last);
  qsort_xrat(p,xrat,left,last-1);
  qsort_xrat(p,xrat,last+1,right);
} /* qsort_xrat */

int ck_for_swap(struct p_data *p,X_ratios *xrat,
	     int left,int right)
   /* return 1 if left wins, else 0 */
{
  double cmpl,cmpr;
  complex HEX_X={-.5,.8660254040};
  X_ratios *tmp;

  if (left==right) return 0;
  if (order_code==1) { /* imaginary part more negative? */
    if (xrat[right].x_ratio.im>xrat[left].x_ratio.im) 
      return 1;
    return 0;
  }
  if (order_code==2) { /* largest ratio of radii (>1) */
    if ((cmpl=p->packR_ptr[xrat[left].v1].rad/
	 p->packR_ptr[xrat[left].v2].rad)<1)
      cmpl=1/cmpl;
    if ((cmpr=p->packR_ptr[xrat[right].v1].rad/
	 p->packR_ptr[xrat[right].v2].rad)<1)
      cmpr=1/cmpr;
    if (cmpr<cmpl) return 1;
    return 0;
  }
  if (order_code==3) { /* furthest from -1/2+sqrt(3)/2 i */
    cmpr=cAbs(csub(xrat[right].x_ratio,HEX_X));
    cmpl=cAbs(csub(xrat[left].x_ratio,HEX_X));
    if (cmpr<cmpl) return 1;
    return 0;
  }
  return 0;
} /* ck_for_swap */

void swap_xrat(X_ratios *xrat,int i,int j)
{
  X_ratios *tmp=NULL;

  tmp=(X_ratios *)calloc(1,sizeof(X_ratios));
  *tmp=xrat[i];
  xrat[i]=xrat[j];
  xrat[j]=*tmp;
  free(tmp);
} /* swap_xrat */

	   
    
