#include "sa_types.h"
#include "sa_proto.h"
#include <math.h>

/* Routines for checking validity of data stored for a packing.
(Add new checks as needs arises.) Check for:
* illegal/inf radii
* negative/nan/inf angle sums,
* nan/inf centers,
* nan/inf aims,
* negative/nan/inf overlaps

Store lists in named file and print message with count of errors. 
Return 0 if everything okay; otherwise return count of errors. */

int nan_check(struct p_data *p,char *datastr)
{
  int v,count=0,prob=0;
  char filename[128],buff[128];
  struct R_data *pR_ptr=p->packR_ptr;
  FILE *fperr=NULL;

  stripsp(datastr);
  if (!p->status) return 0;
  if (!strlen(datastr)) strcpy(buff,"nan"); /* default name */
  else sscanf(datastr,"%100s",buff);
  sprintf(filename,"/tmp/%s_%d.txt",buff,(int)getpid());
  if (!(fperr=fopen(filename,"w"))) return 0;
  fprintf(fperr,"CP nan_check file, hes = %d, nodecount = %d.\n\n",
	  p->hes,p->nodecount);

  sprintf(msgbuf,"nan_check (see file %s for details): \n",filename);
  msg();

  /* illegal/nan/inf radii */
  fprintf(fperr,"Radii problems: \n  ");
  count=0;
  if (p->hes>0) /* spherical */
    {
      for (v=1;v<=p->nodecount;v++)
	if (isnan((singledouble)pR_ptr[v].rad) 
	    || isinf((singledouble)pR_ptr[v].rad)
	    || pR_ptr[v].rad <=0 || pR_ptr[v].rad>=M_PI) 
	  {
	    fprintf(fperr,"%d ",v);
	    count++;
	  }
    }
  else if (p->hes<0) /* hyperbolic (s_radii) */
    {
      for (v=1;v<=p->nodecount;v++)
	if (isnan((singledouble)pR_ptr[v].rad) 
	    || pR_ptr[v].rad>=1.0
	    || (!p->packK_ptr[v].bdry_flag && pR_ptr[v].rad <=0))
	  {
	    fprintf(fperr,"%d ",v);
	    count++;
	  }
    }
  else /* euclidean */
    {
      for (v=1;v<=p->nodecount;v++)
	if (isnan((singledouble)pR_ptr[v].rad) 
	    || isinf((singledouble)pR_ptr[v].rad)
	    || pR_ptr[v].rad <=0)
	  {
	    fprintf(fperr,"%d ",v);
	    count++;
	  }
    }
  if (!count) 
    {
      fprintf(fperr,"none\n");
      sprintf(msgbuf,"  No radius problems.");
    }
  else
    {
      prob++;
      fprintf(fperr,"; count = %d\n",count);
      sprintf(msgbuf,"  %d problem radii found.\n",count);
    }
  msg();

  /* illegal/nan/inf centers */
  fprintf(fperr,"Circle center problems: \n ");
  count=0;
  if (p->hes>0) /* spherical */
    {
      for (v=1;v<=p->nodecount;v++)
	if (isnan((singledouble)pR_ptr[v].center.re) 
	    || isnan((singledouble)pR_ptr[v].center.im)
	    || isinf((singledouble)pR_ptr[v].center.re)
	    || isinf((singledouble)pR_ptr[v].center.im)
	    || pR_ptr[v].center.im<0 || pR_ptr[v].center.im>M_PI+toler)
	  {
	    fprintf(fperr,"%d ",v);
	    count++;
	  }
    }
  else if (p->hes<0) /* hyperbolic */
    {
      for (v=1;v<=p->nodecount;v++)
	if (isnan((singledouble)pR_ptr[v].center.re) 
	    || isnan((singledouble)pR_ptr[v].center.im)
	    || isinf((singledouble)pR_ptr[v].center.re)
	    || isinf((singledouble)pR_ptr[v].center.im)
	    || cAbs(pR_ptr[v].center)>1.0+toler)
	  {
	    fprintf(fperr,"%d ",v);
	    count++;
	  }
    }
  else /* euclidean */
    {
      for (v=1;v<=p->nodecount;v++)
	if (isnan((singledouble)pR_ptr[v].center.re) 
	    || isnan((singledouble)pR_ptr[v].center.im)
	    || isinf((singledouble)pR_ptr[v].center.re)
	    || isinf((singledouble)pR_ptr[v].center.im))
	  {
	    fprintf(fperr,"%d ",v);
	    count++;
	  }
    }
  if (!count) 
    {
      fprintf(fperr,"none\n");
      sprintf(msgbuf,"  No center problems.");
    }
  else
    {
      prob++;
      fprintf(fperr,"; count = %d\n",count);
      sprintf(msgbuf,"  %d problem centers found.\n",count);
    }
  msg();

  /* illegal/nan/inf angle_sums */
  fprintf(fperr,"Angle sum problems: \n  ");
  count=0;
  for (v=1;v<=p->nodecount;v++)
    if (isnan((singledouble)pR_ptr[v].curv))
      {
	fprintf(fperr,"%d ",v);
	count++;
      }
  if (!count) 
    {
      fprintf(fperr,"none\n");
      sprintf(msgbuf,"  No angle sum problems.");
    }
  else
    {
      prob++;
      fprintf(fperr,"; count = %d\n",count);
      sprintf(msgbuf,"  %d problem angle sums found.\n",count);
    }
  msg();

  /* fixup: add checks of other things later */

  fprintf(fperr,"\n END");
  fclose(fperr);
  return prob;
} /* nan_check */

  
  
