#include "sa_types.h"
#include "sa_proto.h"

using namespace std;

// We use fstreams instead of FILE pointers in C++.  The
// leading '&' in the 1st parameter to writepath indicates
// pass-by-reference, which is similar to using a pointer
// except that the caller does not need to take the address
// of the variable.  I.e. to open a file for writing, we
// would create an ofstream (ofstream is a derived class of
// fstream)  and pass it to writepath as follows:
//
// ofstream myFile("filename");
// writepath(myFile, pathlist);
//
// Note:  If you want to create a file for both reading and
// writing, you should use an object of type fstream rather
// than type ofstream:
//
// fstream myFile("filename", ios::in | ios::out);
//
// The ios:: prefix to the bit flags in the fstream constructor
// indicate that the flags are defined as constants by the
// fstream class.
//
// There are two primary reasons for using pass by reference:
// + It is more efficient
// + We don't know (and shouldn't) how the operations on an
//   IO object change its internal state, and whether those
//   changes should be saved.  Passing by reference ensures
//   that any changes to the underlying object are saved.

int new_writepath(ofstream &outfile, struct Pathlist *pathlist)

{
    Pathlist *trace;
    double x, y;
    int count;
    // This code is not completely equivalent to the above.
    // For illustration purposes, additional formatting has been added.
    // The output is similar to fprintf(fp, "%38.30f", floatvar), assuming
    // MAX_ACCUR is 30.

    // This tests if the stream is in an error state or at EOF

    if(outfile.fail() || outfile.eof())  return 0;

    // Use scientific notation.  The 2nd parameter clears any previous
    // formatting specifier for displaying floats.

    outfile.setf(ios::scientific, ios::floatfield);

    // Use right justification.  The 2nd parameter clears any previous
    // justification setting.

    outfile.setf(ios::right, ios::adjustfield);

    // Sets the precision to MAX_ACCUR digits.  Unlike the scientific
    // notation/justification specifiers, it is not necessary to clear
    // the previous precision value; this is done automatically.

    outfile.precision(MAX_ACCUR);

    // The general syntax for writing to an output stream is
    //
    // streamname << data_item_1 [ << data_item_2 [ ... << data_item_n ] ];

    outfile << "PATH\n";
    trace = pathlist;
    count = 0;
    while(trace!=NULL)  {
        x = trace->x;
        y = trace->y;

        // The setw IO manipulator sets the field width.  This must be called
        // before every piece of data you push on the stream, or the default
        // field width is used.  IO manipulators get pushed on an IO stream
        // similarly to data, but serve to change formatting of the stream
        // rather than actually providing data.

        outfile << setw(MAX_ACCUR + 8) << x << setw(MAX_ACCUR + 8) << y << "\n";
        trace=trace->next;
        count++;
    }
    outfile << "END\n";

    // One last note:
    // Some people like to end their lines with the endl constant when using
    // C++ IO, as in
    //
    // outfile << "some string" << endl;
    //
    // instead of
    //
    // outfile << "some string\n";
    //
    // However, I recently found out this is a bad practice for disk files,
    // since endl flushes the buffer in addition to ending the line.

    return count;
} /* new_writepath */

/*
// Other useful output functions not shown above:
//
// char ch;
// int len = 256;
// char buf[len];
//
// infile.put(ch);            // Like putc(infile, fp);
// infile.write(buf, len);    // Like fwrite()
*/
