#include "sa_types.h"
#include "sa_proto.h"

/* Measures of packing quality to help identify layout errors,
misplaced circles, poorly packed faces, etc. Started 4/2002.
Might find better organization in future.

face_error:  sets plot_flag (<=0 means poorly plotted) of faces
based on whether their plotted circles fit as intended.

pin_face:  returns the index of face's vertex most reliably placed;
to help fix the face. 

*/

int face_error(struct p_data *p,double crit,struct Vertlist *facelist)
     /* Compare e = edgelength of faces (based on circle computed
centers) to d = desired edgelength (based on radii and overlaps).
Those in facelist with 1/crit < (e/d) < crit have plot_flag set to 0. 
Crit should be greater than 1; calling routine cleans up facelist.
Return -1 on error or count of poorly plotted faces. (User may want to
set all plot_flags before calling this routine, eg. 'set_plot_flags'.) */
{
  int i,count=0,v1,v2,face;
  double elength,dlength,recip,quo;
  struct Vertlist *trace;
  
  if (crit<=1.0) return -1;
  recip=1/crit;
  trace=facelist;
  while (trace && p->faces[(face=trace->v)].plot_flag>0)
    {
      for (i=0;i<3;i++)
	{
	  v1=p->faces[face].vert[i];
	  v2=p->faces[face].vert[(i+1)%3];
	  if ((elength=edge_length(p,v1,v2))>=0.0
	      && (dlength=desired_length(p,v1,v2))>0.0
	      && ((quo=elength/dlength)>crit || quo<recip))
	    {
	      p->faces[face].plot_flag=0;
	      count++;
	      i=3; /* break out of for loop */
	    }
	} /* end of for loop */
      trace=trace->next;
    } /* end of while */
  return count;
} /* face_error */
	  
int pin_face(struct p_data *p,int face)
     /* Criterion: percentage of petals within twice the 
	desired distance. */
{
  int i,j,v,w,qual[3],it=0,num;
  double dlength,elength,newrat,rat=-1.0;
  struct K_data *pK_ptr=p->packK_ptr;

  for (i=0;i<3;i++)
    {
      qual[i]=0;
      v=p->faces[face].vert[0];
      num=pK_ptr[v].num+pK_ptr[v].bdry_flag;
      for (j=0;j<num;j++)
	{
	  w=pK_ptr[v].flower[j];
	  if ((dlength=desired_length(p,v,w))>0.0
	      && (elength=edge_length(p,v,w))>0.0
	      && elength<2.0*dlength)
	    qual[i] += 1;
	}
      if ((newrat=((double)qual[i])/((double)num))>rat)
	{
	  rat=newrat;
	  it=v;
	}
    }
  return it;
} /* pin_face */
  

double desired_length(struct p_data *p,int v,int w)
     /* return intended length of edge based on radii and overlap.
      Not ready for sph case. Return -1 on error. */
{
  int indx;
  struct R_data *pR_ptr=p->packR_ptr;

  if (p->hes>0) /* sph case; fixup for overlap */
    return (pR_ptr[v].rad+pR_ptr[w].rad);
  else if (p->hes==0) /* eucl */
    {
      if (p->overlap_status && (indx=nghb(p,v,w))>=0)
	return (sqrt(pR_ptr[v].rad*pR_ptr[v].rad
		     +pR_ptr[w].rad*pR_ptr[w].rad
		     +2.0*pR_ptr[v].rad*pR_ptr[w].rad
		     *p->packK_ptr[v].overlaps[indx]));
                             /* fixup: is this formula right? */
      return (pR_ptr[v].rad+pR_ptr[w].rad);
    }
  else /* hyp */
    {
      if (pR_ptr[v].rad<=0 || pR_ptr[w].rad<=0) return -1;
      if (p->overlap_status && (indx=nghb(p,v,w))>=0)
	return (h_invdist_length(pR_ptr[v].rad,pR_ptr[w].rad,
			      p->packK_ptr[v].overlaps[indx]));
      else
	return (radius(p,v)+radius(p,w));
    }
} /* desired_length */

double edge_length(struct p_data *p,int v,int w)
     /* return distance between points; -1 on error */
{
  complex a,b;

  a=p->packR_ptr[v].center;
  b=p->packR_ptr[w].center;
  if (p->hes>0) /* sph */
    return (s_dist(a,b));
  else if (!p->hes) /* eucl */
    return (cAbs(csub(a,b)));
  if (cAbs(a)>=1.0 || cAbs(b)>=1.0) return -1;
  return h_dist(a,b,0);
} /* edge_length */
      
