/* random_packing.cpp Original code by Gerald Orick, 2006. */

#include "sa_head.h"
#include "sa_types.h"
#include <sys/time.h>

typedef struct {
  // links are stored in ccw fashion
  int flags;
  int left_index; // note indexes are mid-cursor points
  int right_index;
  int link_count;
  int link_allocation;
  int *link_list;
} node_struct;

typedef struct {
  int node_count;
  int free_node_count;
  int next_free_node;
  node_struct *nodes;
} mesh_struct;

const int MROOT_LINK_FLAG = -1;
const int MBORDER_FLAG = 1;
const int MINTERIOR_FLAG = 0;

const int DEFAULT_LINK_ALLOCATION = 2*4;

// constants for distribution

#define M_ADD 1
#define M_JOIN 2
#define M_SKIP 3

int distribution();
int next_border_node(mesh_struct *msp,int n);
int grow_link_list(node_struct *nsp);
void left_link_insert(node_struct *nsp,int n);
void right_link_insert(node_struct *nsp,int n);
int insert_border_node(mesh_struct *msp,int node);
int check_for_link(node_struct *nsp,int n);
int join_border_node(mesh_struct *msp,int node);
mesh_struct *new_mesh(int n);
void make_random_mesh(mesh_struct *msp,int size);
void finalize_random_mesh(mesh_struct *msp,int n);
int output_random_mesh(struct p_data *p,mesh_struct *msp);
void output_text (mesh_struct *msp);

// local procedures, 
int random_pack_call(struct p_data *p,char *datastr) {
  int n;
  
  if (!sscanf(datastr,"%d",&n) || n<4) return 0;
  mesh_struct *msp;
  if (!(msp=new_mesh(n))) return 0;
  make_random_mesh(msp,n);
  finalize_random_mesh(msp,n);
  return output_random_mesh(p,msp);
//  output_text(msp);
//  return 1;
} /* random_pack_call */

mesh_struct *new_mesh(int n) {
  int k; // utility

  // make storage and setup data
  mesh_struct *msp;
  msp = (mesh_struct *)malloc(sizeof(mesh_struct));
  if (msp == NULL) { //allocation error
    return NULL;
  }

  msp->nodes = (node_struct *)malloc(n * sizeof(node_struct));
  if (msp->nodes == NULL) { //allocation error
    free(msp);
    return NULL;
  }
  msp->node_count =0;
  msp->next_free_node = 0;
  msp->free_node_count = n;
  
  // now set up each link list
  
  for (k=0; k<n; k++) {
    node_struct *nsp;
    nsp = &(msp->nodes[k]);
    nsp->link_list = (int *)malloc(sizeof(int) * (DEFAULT_LINK_ALLOCATION+1));
    if (nsp->link_list == NULL) {
    	int j;
    	for (j=0;j<k;j++) free(msp->nodes[j].link_list);
    	free(msp->nodes);
		free(msp);// allocation error
	    return NULL;
    }
    nsp->flags = nsp->link_count = 0; 
    nsp->link_allocation = DEFAULT_LINK_ALLOCATION;
    nsp->left_index = nsp->right_index = DEFAULT_LINK_ALLOCATION/2;
  }

  // always start with three vertices
  for (k=0; k<3; k++) {
    msp->nodes[k].flags = MBORDER_FLAG;
  };

  left_link_insert(&(msp->nodes[0]), 1);
  right_link_insert(&(msp->nodes[0]), 2);
  left_link_insert(&(msp->nodes[1]), 2);
  right_link_insert(&(msp->nodes[1]), 0);
  left_link_insert(&(msp->nodes[2]), 0);
  right_link_insert(&(msp->nodes[2]), 1);
 
  msp->node_count = 3;
  msp->next_free_node = 3;
  msp->free_node_count -= 3;

  return(msp);
} /* new_mesh */

void make_random_mesh(mesh_struct *msp,int size) {
  // have already initialized size of objective
  int c_border_node = 0;
  int mesh_action;

  while (msp->node_count < size) 
    {
      int save_node; // used for debug return info
      mesh_action = distribution();
      
      if (msp->node_count <=3) mesh_action = M_ADD;
      // prevent (*msp) from getting too small
      switch(mesh_action) {
      case M_SKIP: // M_SKIP 
	save_node = c_border_node;
	c_border_node = next_border_node(msp, c_border_node); 
	//fprintf(stderr, "SKIP %ld\n", save_node);
	break;
      case M_ADD: // M_ADD 
	save_node = c_border_node;
	c_border_node = insert_border_node(msp, c_border_node);
	//fprintf(stderr, "ADD %ld\n", save_node);
	break;
      case M_JOIN: // M_JOIN
	save_node = c_border_node;
	c_border_node = join_border_node(msp, c_border_node); 
	//fprintf(stderr, "JOIN %ld\n", save_node);
	break;
      }
    }
} /* make_random_mesh */

void finalize_random_mesh(mesh_struct *msp,int n) {
  /* here get rid of undesired vertices, etc. */
  
  // One pass around the border to add edges between any border 
  //   vert in just one face and the nextnext border vert
  int count=0; // safty measure to avoid infinite loop
  int wflag=0; // to get through the loop the first time
  int start_indx; // know where we started 
  int next_indx;
  int k=0;
  while(k<msp->node_count && msp->nodes[k].flags==MINTERIOR_FLAG) k++;
  if (k==msp->node_count) return; // didn't find any boundary node (error)
  next_indx=start_indx=k;
  while ((!(wflag++) || next_indx!=start_indx) && count < msp->node_count) { // go around boundary
  	if (msp->nodes[next_indx].link_count==2) 
  		join_border_node(msp,next_indx);
  	next_indx=next_border_node(msp,next_indx);
  	count++;  // safty measure to avoid infinite loop
  }
} /* finalize_random_mesh */

int output_random_mesh(struct p_data *p,mesh_struct *msp) {
  /* generate the packing with the resulting combinatorics
     and max pack it */
  int *newflower=NULL;
  empty_pack(p);
  alloc_pack_space(p,msp->node_count+1,1);
  struct K_data *pK_ptr=p->packK_ptr;
  struct R_data *pR_ptr=p->packR_ptr;

  p->nodecount=msp->node_count;
  p->status=1;
  p->hes=-1; // make it hyperbolic
  // loop through the nodes, reading off the links
  int i,j,k;
  for (i=0;i<msp->node_count; i++) {
    int faces;
    faces = msp->nodes[i].link_count;
    if (msp->nodes[i].flags == MBORDER_FLAG) faces--;
    newflower=(int *)calloc((size_t)(faces+1),sizeof(int));
    int save_left_index = msp->nodes[i].left_index;
    j=0;
    for (k=msp->nodes[i].left_index; k < msp->nodes[i].right_index; k++) {
      newflower[j++]=msp->nodes[i].link_list[k]+1;
    }
    if (msp->nodes[i].flags != MBORDER_FLAG) { // close up
      newflower[j]=newflower[0];
      pK_ptr[i+1].bdry_flag=0;
    }
    else pK_ptr[i+1].bdry_flag=1;
    pK_ptr[i+1].num=faces;
    pK_ptr[i+1].flower=newflower;
  }

  // fix up the packing
  for (i=1;i<=p->nodecount;i++) pR_ptr[i].rad=.5;
  complex_count(p,1);
  choose_alpha(p);choose_beta(p);choose_gamma(p);
  facedraworder(p,0);
  p->active_node=p->alpha;
  sprintf(p->file_name,"random_%d.p",p->nodecount);
  new_pack_setup(p);

// max_pack
  int dum;
  int pnum=pack_num(p);
  sprintf(buf,"max_pack -p%d",pnum);
  handle_cmd(buf,&dum);
} /* output_random_mesh_text */

void left_link_insert (node_struct *nsp,int n) {
  if (nsp->left_index > 0) { // OK to insert
    nsp->left_index--;
    nsp->link_list[nsp->left_index] = n;
    nsp->link_count++;
  }
  else {
    grow_link_list(nsp);
    left_link_insert(nsp, n);
  }
} /* left_link_insert */

void right_link_insert (node_struct *nsp,int n) {
  if (nsp->right_index < nsp->link_allocation) { // OK to insert
    nsp->link_list[nsp->right_index] = n;
    nsp->right_index++;
    nsp->link_count++;
  }
  else {
    grow_link_list(nsp);
    right_link_insert(nsp, n);
  }
} /* right_link_insert */

int grow_link_list(node_struct *nsp) {
  int newsize, leading_count;
  newsize = nsp->link_allocation + DEFAULT_LINK_ALLOCATION;
  leading_count = DEFAULT_LINK_ALLOCATION / 2;

  int *newlink_list;
  newlink_list = (int *)malloc(sizeof(int) * (newsize+1));
  if (newlink_list == NULL) { //allocation error
    return 0;
  }
  
  int k;
  for (k=nsp->left_index; k<nsp->right_index; k++) 
    newlink_list[k + leading_count] = nsp->link_list[k];
  nsp->left_index += leading_count;
  nsp->right_index += leading_count;
  nsp->link_allocation = newsize;
  free(nsp->link_list);
  nsp->link_list = newlink_list;
  return 1;
} /* grow_link_list */

int distribution()
{
  // This routine determines which action choices are SKIP ADD JOIN
  // need to add multiple operations and directional (left/right) to 
  // make more symmetric
  int r_val = random();
  int step = (RAND_MAX) / 3;
  return (r_val/step + 1);
} /* distribution */

int insert_border_node(mesh_struct *msp,int node) {
  // newnode // dual links // be sure to mark border

  // newnode
  int newid;
  newid = msp->next_free_node++; 
  msp->free_node_count--;
  msp->node_count++;
  if (msp->free_node_count < 0) { // allocation error
    return 0;
  }
  node_struct *nsp;
  nsp = &(msp->nodes[newid]);
  nsp->left_index  = nsp->right_index = DEFAULT_LINK_ALLOCATION / 2;
  nsp->flags = MBORDER_FLAG;

  // now connect the dots
  int nextnode;
  nextnode = next_border_node(msp, node);
  left_link_insert(nsp, nextnode);
  right_link_insert(nsp, node);
  right_link_insert( &(msp->nodes[nextnode]), newid);
  left_link_insert( &(msp->nodes[node]), newid);
  return (newid);
} /* insert_border_node */

int next_border_node(mesh_struct *msp,int n) {
  int *nodelinks = msp->nodes[n].link_list;
  return nodelinks[msp->nodes[n].left_index];
} /* next_border_node */

int check_for_link(node_struct *nsp,int n) {
  int status = 0;
  int k;
  for (k=nsp->left_index; k<nsp->right_index; k++) 
    status |= (nsp->link_list[k]==n);

  return (status);
} /* check_for_link */

int join_border_node(mesh_struct *msp,int node) {
  // sink node // make the links // 
  int nextnode; 
  int nextnextnode;
  nextnode = next_border_node(msp, node);
  nextnextnode = next_border_node(msp, nextnode);
  if (!check_for_link( &(msp->nodes[node]), nextnextnode)) { //we are OK
    right_link_insert(&(msp->nodes[nextnextnode]), node);
    left_link_insert(&(msp->nodes[node]), nextnextnode);
    msp->nodes[nextnode].flags = MINTERIOR_FLAG;
  }
  return (nextnextnode);
} /* join_border_node */

void output_text ( mesh_struct *msp) {
  printf("NODECOUNT:\t%ld\n\n", (*msp).node_count);
  printf("GEOMETRY:\teuclidean\n\n");
  printf("FLOWERS:\n\n");

  // loop through the nodes
  // list the links
  int i,k;
  for (i=0; i< (*msp).node_count; i++) {
    int faces;
    faces = (*msp).nodes[i].link_count;
    if ((*msp).nodes[i].flags == MBORDER_FLAG) faces--;
    printf ("%ld\t%ld\t",i+1, faces);
    

    int save_left_index = (*msp).nodes[i].left_index;
    for (k=(*msp).nodes[i].left_index; k < (*msp).nodes[i].right_index; k++) {
      printf("%ld ", (*msp).nodes[i].link_list[k]+1);
    }
    if (!(*msp).nodes[i].flags == MBORDER_FLAG)
      printf("%ld ", (*msp).nodes[i].link_list[save_left_index]+1);
    printf("\n");

  }
  printf("\nEND\n");
}


