#include "sa_types.h"
#include "sa_proto.h"

/* 7/2000

Open a file (*.vl) containing a parent packing (optional, occurs 
near beginning, after some keywords) and a sequence of lists of 
vertices to be the 'variable' vertices of associated child 'light' 
packings. 'parentpacking' is created elsewhere, by the calling 
program or embedded in the *.vl file.

If PARENT is given, then allocate the packing space and read it.
Else, check if 'parentpacking' is given and CHECK_COUNT matches its
nodecount, and use that as the parent. Allocate the light packing 
space, read in the lists, create the light packings.

Idea is to create these data files using various decompositions
to specify decompositions into light packings, generally as input 
to various repacking routines. 

Return 0 on error or n, number of light packings.

Input file format (at least for developmental purposes):
("[..]" data was introduced for overlaps with mpi runs.)

   CHECK_COUNT: {nodecount of parent}
   VERT_LIST_COUNT: n (limit n=200. optional: else read up to 200 lists.)

   PARENT: {parent packing}

   [NUM_GENERATIONS: {num_generations}] (for adding overlap buffers)

   VERT_LIST: {list-1} (done)
   [WEIGHT_LIST: {weight_list-1} (done)] (depth of verts from bdry)
   ..
   ..
   VERT_LIST: {list-n} (done)
   [WEIGHT_LIST: {weight_list-n} (done)]

 */

int read_parent_child(char *filename,struct p_data *parentpacking,
		      struct p_light ***plight, int *overlap)
{
  int i,j,v,N,checkcount,w;
  long ftel;
  char buf[1024],command[64];
  struct p_data *p=parentpacking;
  struct K_data *pK_ptr;
  struct p_light **pl_ptr=NULL;
  FILE *fp=NULL;
  int *vertlist;

  *overlap = 0;

  if (!parentpacking || !(fp=fopen(filename,"r"))) 
    {
      printf("parentpacking = %p\nfp = %p\n", parentpacking, fp);
      sprintf(msgbuf,"parentpacking space not allocated or failure "
	      "to open file %s.",filename);
      emsg();
      goto BOMB;
    }

  /* must find CHECK_COUNT: and VERT_LIST_COUNT: first */
  if (!fscanf(fp,"%127s",buf) || (strcmp(buf,"CHECK_COUNT:"))
      || !fscanf(fp,"%d",&checkcount) || checkcount<4 
      || !fscanf(fp,"%127s",buf) || (strcmp(buf,"VERT_LIST_COUNT:")))
    {
      sprintf(msgbuf,"A parent/child *.vl file must begin with "
	      "'CHECK_COUNT:' and 'VERT_LIST_COUNT:'.");
      emsg();
      goto BOMB;
    }
  if (!fscanf(fp,"%d",&N) || N<1 || N>200) N=200; 

  /* if PARENT is included and nodecount matches CHECK_COUNT, then
     read that as the parent packing. Otherwise, use given parentpacking 
     if its nodecount matches CHECK_COUNT. Otherwise, error. */
  
  ftel=ftell(fp);
  if (fscanf(fp,"%127s",buf) && !(strcmp(buf,"PARENT:")))
    {
      if (!alloc_pack_space(p,10000,0) || !readpack(fp,p))
	/* if PARENT is indicated but reading it fails, error */
	{
	  sprintf(msgbuf,"Failed in reading given parent or "
		  "allocating space.");
	  emsg();
	  goto BOMB;
	}
    }
  else /* didn't have packing in data, use parentpacking */
    {
      fseek(fp,ftel,SEEK_SET); /* reset file position */
      if (p->nodecount!=checkcount)
	{
	  sprintf(msgbuf,"No packing given in data or as arguement.");
	  emsg();
	  goto BOMB;
	}
    }
  pK_ptr=p->packK_ptr;

  /* now create light packing space and read vert lists */
  if (!(pl_ptr=(struct p_light **)calloc(N+1,sizeof(struct p_light *))))
    {
      sprintf(msgbuf,"Failed to allocate space for %d light packings.",
	      N);
      emsg();
      goto BOMB;
    }
  
  vertlist = (int *) calloc(p->nodecount + 1, sizeof(int));
  i=1;
  while (i<=N)
    {
      while((fscanf(fp,"%63s",command) != EOF))  {
        if(strcmp(command, "NUM_GENERATIONS:") == 0)  {
            if(fscanf(fp, "%d", overlap) != 1)  {
                sprintf(msgbuf, "invalid overlap value\n");
                emsg();
                goto BOMB;
            }
        }
        else if(strcmp(command, "VERT_LIST:") == 0 
		|| strcmp(command, "WEIGHT_LIST:") == 0){
          break;
        }
      }
      if (!strcmp(command,"VERT_LIST:")) /* found vert_list */
	{
	  for (j=1;j<=p->nodecount;j++) pK_ptr[j].mark=0;
          j = 0;
          memset(vertlist, 0, sizeof(int) * (p->nodecount + 1));
	  while ((fscanf(fp,"%63s",command)!=EOF)
		 && sscanf(command,"%d",&v) && v>0 && v<=p->nodecount){
	    if (!pK_ptr[v].mark){
	      pK_ptr[v].mark=1;	  
	      vertlist[j] = v;
	      j++;
	    }
          }
	  if (!(pl_ptr[i]=convert_to_p_light(p)))
	    {
	      sprintf(msgbuf,
		      "Failed to convert list #%d to light packing.",i);
	      emsg();
	      goto BOMB;
	    }
          if((pl_ptr[i]->generation = 
	      (int *)calloc(pl_ptr[i]->counts[0] + 1, sizeof(int))) == NULL){
              sprintf(msgbuf, "memory allocation error\n");
              emsg();
              goto BOMB;
          }
          for(j = 1; j <= pl_ptr[i]->counts[0]; j++)  pl_ptr[i]->generation[j] = 2 * *overlap + 1;
	  i++;
	}
      else if(strcmp(command, "WEIGHT_LIST:") == 0)  {
        if(i <= 1 || pl_ptr[i - 1] == NULL)  {
          sprintf(msgbuf, "WEIGHT_LIST found without a VERT_LIST\n");
          emsg();
          goto BOMB;
        }
        j = 0;
        while((fscanf(fp, "%63s", command) != EOF) && sscanf(command, "%d", &w))  {
          pl_ptr[i - 1]->generation[j] = w;
          j++;
        }
      }
      else
	{
	  sprintf(msgbuf,"Last list found was #%d.",i-1);
	  msg();
	  if(fp) fclose(fp);
	  fp=NULL;
	  *plight=pl_ptr;
          free(vertlist);
	  return i-1;
	}
    } /* end of while */
  sprintf(msgbuf,"Last list found was #%d.",N);
  msg();
  if(fp) fclose(fp);
  *plight=pl_ptr;
  free(vertlist);
  return N;
    
 BOMB:
  if (fp) fclose(fp);
  for (i=1;i<=N;i++)
    if (pl_ptr[i]) free_p_light(pl_ptr+i);
  if (pl_ptr) free(pl_ptr);
  free(vertlist);
  return 0;
} /* read_parent_child */
