/* sa_XML.cpp coordinate graphical setting with java GUI. */

#include "sa_head.h"
#include "Interface.h"

/* I'm trying to isolate calls to the front end from the 
 * general circle packing library stuff. So when output or
 * messages are needed, I call things through the procedures
 * here rather than directly; hopefully, only this file will
 * need Interface.h (eventually).
 * 
 * Changes in graphical settings have to be coordinated with GUI.
 * Ultimately, the user sets things from GUI, but settings often
 * have to be stored here for CPack to work correctly, so there
 * are update issues.
 * 
 * GUI keeps track of:
 * 	 linethickness, opacities, axes mode, initial default settings
 *   of pretty much everything.
 * 
 * CPack keeps track of:
 *   screen size, display mode (ie. filled/empty, circles/faces), 
 *   geometry, packName, spherical viewpoint
 * 
 * CPack needs to know:
 *   linethickness (for use with PostScript output)
 * 
 * Have to adjust things in a dance -- some actions initiated by
 * GUI, some by CPack. So commands to CPack will make changes to
 * things maintained by java, etc.
 */

/* After creating a seed or reading in a packing, have to set
 * GUI stuff. (Note: this will be different if copying a packing;
 * settings are usually just transferred.
 * 
 */ 
int new_pack_setup(struct p_data *p) 
{
 	int pnum=pack_num(p);
 	
 	resetPack(pnum); // resets defaults: antialias, line size, etc.
 	clearCanvas(pnum);
 	setPackName(pnum,p->file_name);
 	updateWorld(pnum);
 	setGeometry(pnum,p->hes);
 	return 1;
} /* new_pack_setup */

int sa_send_flag_opts(struct p_data *p)
{
	setFlagOptions(pack_num(p),p->screen->cflag_opt,p->screen->fflag_opt);
	return 1;
}

/* After switching geometry of a packing, certain things need
 * updating, others stay the same.
 */
int switch_geom(struct p_data *p) 
{
	int pnum=pack_num(p);
	setGeometry(pnum,p->hes);
	return 1;
} /* switch_geom */	

/* fixup: need to re-implement the "stretch" for horocycles
   which shifts their labels for easier reading. Here's the
   old code:

      stretch=1.0+(p->screen->box.rx-p->screen->box.lx)*15/
	(p->screen->pix_box.rx-p->screen->pix_box.lx); 
        ctr.re=p->packR_ptr[v].center.re*stretch;
      ctr.im=p->packR_ptr[v].center.im*stretch;
*/

int canvas_mode_defaults(struct p_data *p)
     /* Set defaults for display mode stuff */
{
  line_thick_mode(p,1);
  if (p->hes>=0) unitcircle_mode(p,0);
  else unitcircle_mode(p,1);
//  send_sph_view(p,0);
  sa_send_geom(p);
  return 1;
} /* canvas_mode_defaults */

int line_thick_mode(struct p_data *p,int t) {
	setLineThick(pack_num(p),t);
	return 1;
} /* line_thick_mode */

int unitcircle_mode(struct p_data *p,int set) {
     /* no longer used with java front end */
    return 1;
} /* unitcircle_mode */

int equator_mode(struct p_data *p,int set) {
     /* no longer used with java front end */
	return 1;
} /* equator_mode */

int sa_send_geom(struct p_data *p) {
	setGeometry(pack_num(p),p->hes);
	return 1;
} /* sa_send_geom */

int set_pack_label(struct p_data *p) {
     /* change a canvas frame title */
	setPackName(pack_num(p),p->file_name);
	return 1;
} /* set_pack_label */

int sa_clear_canvas(struct p_data *p,int flag) {
	clearCanvas(pack_num(p));
	return 1;
} /* sa_clear_canvas */

int send_sph_view(struct p_data *p) {
	setMatrix(p); // send matrix to java
	return 1;
} /* send_sph_view */

int repack_activity_msg(char *datastr)
/* java_fixup: need report packing activity */
{
  return 1;
} /* repack_activity_msg */

int sa_window_bell()
{
  /* java_fixup: tell GUI to ring bell */
  return 1;
} /* sa_window_bell */

int sa_read_script(char *datastr)
/* tell java to load script */
{
	loadScript(datastr);
	return 1;
}

int labellist(struct p_data *p,struct Vertlist *list,int msg_flag,
	      int faces)
/* Send labels for circles or faces (if 'faces' is set) to Java GUI; 
   they can be used on the screen and/or the scratch window. Calling 
   routine must destroy 'list'. msg_flag is 3 by default: bit 1 means
   display on canvas, bit 2 means to put in scratch window. */
/* fixup: should avoid sending labels for display if they are off screen
   unless they are to be put in scratch window. */

{
  int count=0,front,v,pnum=pack_num(p);
  complex ctr;
  struct Vertlist *trace;
  
  if (!list) return 0;
  trace=list;
  while (trace) {
  	v=trace->v;
    if (faces) { // for faces 
    	ctr=face_center(p,v);
    	drawIndex(pnum,ctr.re,ctr.im,v,msg_flag);
    }
    else { // for circles
		ctr=p->packR_ptr[v].center;
    	if (p->hes>0) {
      		ctr=ss_view(p->screen,ctr,1,&front);
    	}
	  	drawIndex(pnum,ctr.re,ctr.im,v,msg_flag);
    }
	trace=trace->next;
    count++;
  } 
  last_index_global=v;
  return count;
} /* labellist */
      
int sa_refresh_canvas(struct p_data *p)
     /* fixup: don't know yet if any action needs to be taken */
{
  return 1;
} /* sa_refresh_canvas */

/* ==================== various message printers ===================== */

int msg()
/* send msgbuf contents to message subwindow */
{
	if (javaMsgHook!=0) { // java must indicate it's ready for messages
		int msgflag=1;
		showMsgBuf(msgflag,msgbuf);
		if (log_fp) fprintf(log_fp,"MSG: %s\n",msgbuf);
		return 1;
	}
	if (log_fp) fprintf(log_fp,"MSG: %s\n",msgbuf);
	return 1;
} /* msg */

int emsg()
/* send emsgbuf contents to message subwindow */
{
	if (javaMsgHook!=0) { // java must indicate it's ready for messages
		int errflag=1;
		showErrorMsgBuf(errflag,emsgbuf);
		if (log_fp) fprintf(log_fp,"ERROR: %s\n",emsgbuf);
	}
	if (log_fp) fprintf(log_fp,"ERROR: %s\n",emsgbuf);
	return 1;
} /* emsg */

int debugmsg()
/* send msgbuf contents to debug message subwindow */
{
	if (javaMsgHook!=0) { // java must indicate it's ready for messages
		int msgflag=1;
		showDebugMsg(msgflag,msgbuf);
		if (log_fp) fprintf(log_fp,"DeBugMSG: %s\n",msgbuf);
		return 1;
	}
	bprintf(msg_out_fp,"debugMsg %s\n",msgbuf);
	bflush(msg_out_fp);
	if (log_fp) fprintf(log_fp,"DeBugMSG: %s\n",msgbuf);
	return 1;
} /* debugmsg */

int smsg()
/* send msgbuf contents to end of scratch subwindow. */
{
	if (javaMsgHook!=0) { // java must indicate it's ready for messages
		int msgflag=1;
		showScratchMsg(msgflag,msgbuf);
		if (log_fp) fprintf(log_fp,"SCRATCH: %s\n",msgbuf);
	}
	if (log_fp) fprintf(log_fp,"SCRATCH: %s\n",msgbuf);
	return 1;
} /* smsg */

	/* message out to remote procedure */

/**
 * Send data on the current canvas view
 */
int send_setViewBox(struct p_data *p) {
	setViewBox(p);
	return 1;
}

/**
 * have Java open "Map" windows with packs p and q. 
 */
int openMap_call(struct p_data *p,struct p_data *q) {
	return openMap(pack_num(p),pack_num(q));
}

int closeMap_call() {
	return closeMap();
}

int quit(int flag)
     /* flag set means to confirm the quit command */
{
  bprintf(disp_fp,"exit CPack");
  bflush(disp_fp);
  exit(0);
  return 1;
} /* quit */

int chg_active_pack(int pnum)
     /* change active pack to pnum */
{
  if (pnum==current_p) return 1;
  live_node[current_p]=live_face[current_p]=0;
  current_p=pnum;
  setActive(current_p);
  return 1;
} /* chg_active_pack */

/**
 * Set the function description in the java "Function" panel.
 */
int set_ftn_text(char *ftn_text) {
	setFtnText(ftn_text);
	return 1;
} /* set_ftn_text */

/**
 * Updating SidePairings: flush the old tools, create any new
 * if they exist (presumably just updated); see calls in 'comp_pack_centers'
 * Return -1 on error.
 */
int updateSidePairs(struct p_data *p) {
	flushSidePairs(p);
	if (p->hes>0 || !p->redfaces) return 1; // haven't yet implemented sph case.
	/* Note: side-pairings from hyp or eucl case can remain valid
	   when packing is converted to spherical, so don't wipe them.*/
	/* MobType is one of:
	 * 0 = general; (maybe) available for any packing, goes in
	 *     mobiusHandler.toolBar.
	 * 1 = sidepairing; for multiply-connected packings, goes in
	 *     sidePairHandler.toolBar.
	 * 2 = generic; available for this packing only, goes in
	 *     utilityHandler.toolBar. */
	int MobType=1;
	char name[2];
	int indx=1; // indices start at 1.
	char c='a'; // lower case for a mobius
	char C='A'; // upper case for its mate
	name[1]='\0';
	int hit=0;
	int j;
	int count=0;
	while (p->edge_pair[indx].edge) {
		// determinant too small; probably this mobius wasn't set
		if (cAbs(det(p->edge_pair[indx].mob))<toler); 
		else if ((j=find_mate_index(p,indx))>=indx && hit<27) {
			name[0]=c+hit; 
			setMobius(p,MobType,name,p->edge_pair[indx].mob);
			name[0]=C+hit;
			setMobius(p,MobType,name,p->edge_pair[j].mob);
			hit++;
			count++;
		}
		indx++;
	}
	return count;
} /* updateSidePairs */

/**
 * Retrieve f(z) where f is the function defined in 
 * the java "Function" panel.
 */
complex get_java_ftn_value(complex z) {
	int errflag;
	return gogetFtnValue(z);
//	return returnFtnValue(z,&errflag);
}
