#include "sa_types.h"
#include "sa_proto.h"

/*  ==============  Routines for choosing the color settings ============ */
/* Note that calling program must set color_ramp_size, default
   being COLOR_RAMP; this is the number of colors in the blue-red
   color ramp. mid=color_ramp_size/2 is white, 1 to (mid-1) are dark to
   light blue, mid+1 to 2*mid are light to dark red. 

   When CirclePack is executed with the -colorfile {filename} command
   line option, a ramp size can be set and special color codes can
   be assigned to indices. See the read_rgb_file procedure. For
   each color specified, a saturated version (.75 times the rgb value)
   is also recorded so color circle/faces can have a darker edge
   with same basic color as fill. */

extern int color_ramp_size;
extern char next[];

int radius_col_comp(struct p_data *p)
     /* color gradations based on radii. */
{
  double b,t;
  int i,mid;
  struct R_data *pR_ptr;

  pR_ptr=p->packR_ptr;
  mid=(int)(color_ramp_size/2);
  b=t=pR_ptr[1].rad;
  for (i=2;i<=p->nodecount;i++)
    {
      if (pR_ptr[i].rad>t) t=pR_ptr[i].rad;
      if (pR_ptr[i].rad<b) b=pR_ptr[i].rad;
    }
  if (b<0) b=0.0;
  if (t<=0 || fabs(t-b)/t<.005) /* problem of small variation */
    {
      for (i=1;i<=p->nodecount;i++) p->packK_ptr[i].color=mid;
      return 1;
    }
  if (p->hes<0)
    for (i=1;i<=p->nodecount;i++)
      {
	p->packK_ptr[i].color=
	  1+(int)((mid-2)*(pR_ptr[i].rad-b)/(t-b));
      }
  else if (p->hes==0)
    for (i=1;i<=p->nodecount;i++)
      {
	p->packK_ptr[i].color=
	  1+(int)((mid-2)*(pR_ptr[i].rad-t)/(b-t));
      }
  return 1;
} /* radius_col_comp */

int area_col_comp(struct p_data *p)
     /* color gradations, index 1-19 (blues). */
{
  double b,t,*areas;
  int i,flag=1,mid;
  struct R_data *pR_ptr;

  pR_ptr=p->packR_ptr;
  mid=(int)(color_ramp_size/2);
  t=0.0;
  b=h_area(pR_ptr[p->faces[1].vert[0]].rad,
	   pR_ptr[p->faces[1].vert[1]].rad,
	   pR_ptr[p->faces[1].vert[2]].rad);
  if ( !(areas=(double *)malloc((p->facecount+1)*sizeof(double))) ) 
    return 0;
  if (p->hes<0)
    for (i=1;i<=p->facecount;i++)
      {
	areas[i]=h_area(pR_ptr[p->faces[i].vert[0]].rad,
			pR_ptr[p->faces[i].vert[1]].rad,
			pR_ptr[p->faces[i].vert[2]].rad);
	t=(areas[i]>t) ? areas[i]:t;
	b=(areas[i]<b) ? areas[i]:b;
      }
  else if (p->hes==0)
    for (i=1;i<=p->facecount;i++)
      {
	areas[i]=e_area(pR_ptr[p->faces[i].vert[0]].rad,
			pR_ptr[p->faces[i].vert[1]].rad,
			pR_ptr[p->faces[i].vert[2]].rad);
	t=(areas[i]>t) ? areas[i]:t;
	b=(areas[i]<b) ? areas[i]:b;
      }
  if (b==0.0) flag=0;
  if (t==0 || fabs(t-b)/t<.005)
    for (i=1;i<=p->facecount;i++) p->faces[i].color=mid;
  else for (i=1;i<=p->facecount;i++)
    {
      p->faces[i].color=
	1+(int)((mid-2)*(areas[i]-t)/(b-t));
    }
  free(areas);
  return (flag);
} /* area_col_comp */

int color_qc(struct p_data *p,struct p_data *q,char *datastr,
	     double *kmax)
     /* For map between 2 (eucl) packings, color faces of p in
shades of red based on quasiconformal dilatation of the map. 
datastr contains various options and list of faces. (Note: the
faces should have corresponding vertices, as when p and q are
packings of the same complex, but we only check that the face indices
are legal.) Return 0 on improper data, count otherwise. */
{
  int count=0,mid,i,hits;
  double ceiling=-1.0,*dil=NULL;
  char *nextpoint=NULL,*endptr;
  struct Vertlist *facelist=NULL,*ftrace;
  struct p_data *pq=NULL;

  if (p->hes || q->hes) return 0; /* can only do eucl for now. */
  mid=(int)(color_ramp_size/2);
  nextpoint=datastr;
  stripsp(nextpoint);
  *kmax=1.0;

  /* datastr: 
     -m {k}     gives double k>=1 is ceiling: color code faces
                with dilatation >k with blue, those below with
		linear gradations of red 
     {f ..}     list of faces, default to max of facecounts of p and q.
  */

  while (nextpoint[0]=='-') /* pick off any options */
    {
      nextpoint++;
      switch (*nextpoint)
	{
	case 'm': /* get specified ceiling */
	  {
	    nextpoint++;
	    if (!grab_next(&nextpoint,next) 
		|| !sscanf(next,"%lf",&ceiling))
	      return 0;
	    break;
	  }
	} /* end of switch */
    }
  stripsp(nextpoint);
  if (*nextpoint=='\0') strcpy(nextpoint,"a\0"); /* default to all faces.*/
  pq = (q->facecount<p->facecount) ? q : p; 
  if (!(facelist=Face_link_parse(pq,nextpoint,&endptr,&hits,
		 &Vlist,&Elist,&Flist,&region,pathlist,pathlength)))
    return 0;
  ftrace=facelist;
  while (ftrace) 
    {
      count++;
      ftrace=ftrace->next;
    }
  dil=(double *)calloc((size_t)(count+1),sizeof(double));
  for (i=1;i<=count;i++) dil[i]=1.0; /* default */
  ftrace=facelist;
  i=1;
  while (ftrace) /* compute the dilatations */
    {
      if (!face_dilatation(p,q,ftrace->v,&dil[i++])) 
	dil[i-1]=-1.0;
      ftrace=ftrace->next;
    }
  if (ceiling>1.0) *kmax=ceiling; 
  else /* must compute kmax, the sup of the dilatations */
    {
      for (i=1;i<=count;i++)
	(*kmax) = (dil[i]>(*kmax)) ? dil[i] : (*kmax);
    }
  ftrace=facelist;
  i=1;
  while (ftrace)
    {
      if (dil[i]>0 && dil[i]<=(*kmax))
	p->faces[ftrace->v].color=
	  mid+1+(int)((mid-2)*(dil[i]-1.0)/((*kmax)-1.0));
      else if (dil[i]>0)
	p->faces[ftrace->v].color=2; /* over ceiling? dark blue. */
      else
	p->faces[ftrace->v].color=1; /* couldn't compute? black */
      ftrace=ftrace->next;
      i++;
    }
  if (facelist) vert_free(&facelist);
  if (dil) free(dil);
  return count;
} /* color_qc */

int color_qc_xyz(struct p_data *p,struct xyzPoint *xyz_list,
		 int xyzcount,char *datastr,double *kmax)
     /* For map between xyz points and a (eucl) circle packing,
color faces of p as described in color_qc. xyz_list contains points
indexed from 1 to xyzcount. Return 0 on error. */
{
  int i,j,k,mid,hits,count=0,face,it,n;
  double r1,r2,r3,a,b,c,A,B,C,o1,o2,o3;
  double ceiling=-1.0,*dil=NULL;
  char *nextpoint=NULL,*endptr;
  struct Vertlist *facelist=NULL,*ftrace;
  struct R_data *pR_ptr=p->packR_ptr;
  struct K_data *pK_ptr=p->packK_ptr;

  if (p->hes) { /* only euclidean for now */
    sprintf(msgbuf,"color: quasiconformal coloring only for euclidean "
	    "packings at this time.");
    emsg();
    return 0; 
  }
  mid=(int)(color_ramp_size/2);
  nextpoint=datastr;
  stripsp(nextpoint);
  *kmax=1.0;

  /* datastr: 
     -m {k}     gives double k>=1 is ceiling: color code faces
                with dilatation >k with blue, those below with
		linear gradations of red. 
     {f ..}     list of faces, default to max of facecounts of p and q.
  */

  while (nextpoint[0]=='-') /* pick off any options */
    {
      nextpoint++;
      switch (*nextpoint)
	{
	case 'm': /* get specified ceiling */
	  {
	    nextpoint++;
	    if (!grab_next(&nextpoint,next) 
		|| !sscanf(next,"%lf",&ceiling))
	      return 0;
	    break;
	  }
	} /* end of switch */
    }
  stripsp(nextpoint);
  if (*nextpoint=='\0') strcpy(nextpoint,"a\0"); /* default to all faces.*/
  if (!(facelist=Face_link_parse(p,nextpoint,&endptr,&hits,
		 &Vlist,&Elist,&Flist,&region,pathlist,pathlength)))
    return 0;
  ftrace=facelist;
  while (ftrace) 
    {
      count++;
      ftrace=ftrace->next;
    }
  dil=(double *)calloc((size_t)(count+1),sizeof(double));
  for (it=1;it<=count;it++) dil[it]=1.0; /* default */
  ftrace=facelist;
  it=1;
  while (ftrace) /* compute the dilatations */
    {
      face=ftrace->v;
      if ((i=p->faces[face].vert[0])<=xyzcount
	  && (j=p->faces[face].vert[1])<=xyzcount
	  && (k=p->faces[face].vert[2])<=xyzcount)
	/* yes, we have xyz locations for these */
	{
	  o1=o2=o3=1.0;
	  if (p->overlap_status) {
	    if ((n=nghb(p,i,j))>=0) o1=pK_ptr[i].overlaps[n];
	    if ((n=nghb(p,j,k))>=0) o2=pK_ptr[j].overlaps[n];
	    if ((n=nghb(p,k,i))>=0) o3=pK_ptr[k].overlaps[n];
	  }
	  r1=pR_ptr[i].rad;
	  r2=pR_ptr[j].rad;
	  r3=pR_ptr[k].rad;
	  A=sqrt(o1*r1*r2+r1*r1+r2*r2);
	  B=sqrt(o2*r2*r3+r2*r2+r3*r3);
	  C=sqrt(o3*r3*r1+r3*r3+r1*r1);
	  a=sqrt(
	      (xyz_list[i].x-xyz_list[j].x)*(xyz_list[i].x-xyz_list[j].x)
	      + (xyz_list[i].y-xyz_list[j].y)*(xyz_list[i].y-xyz_list[j].y)
	      + (xyz_list[i].z-xyz_list[j].z)*(xyz_list[i].z-xyz_list[j].z));
	  b=sqrt(
	      (xyz_list[k].x-xyz_list[j].x)*(xyz_list[k].x-xyz_list[j].x)
	      + (xyz_list[k].y-xyz_list[j].y)*(xyz_list[k].y-xyz_list[j].y)
	      + (xyz_list[k].z-xyz_list[j].z)*(xyz_list[k].z-xyz_list[j].z));
	  c=sqrt(
	      (xyz_list[i].x-xyz_list[k].x)*(xyz_list[i].x-xyz_list[k].x)
	      + (xyz_list[i].y-xyz_list[k].y)*(xyz_list[i].y-xyz_list[k].y)
	      + (xyz_list[i].z-xyz_list[k].z)*(xyz_list[i].z-xyz_list[k].z));
	  if (!e_dilatation(A,B,C,a,b,c,&dil[it++])) 
	    dil[it-1]=-1.0;
	}
      ftrace=ftrace->next;
    }
  if (ceiling>1.0) *kmax=ceiling; 
  else /* must compute kmax, the sup of the dilatations */
    {
      for (it=1;it<=count;it++)
	(*kmax) = (dil[it]>(*kmax)) ? dil[it] : (*kmax);
    }
  ftrace=facelist;
  it=1;
  while (ftrace)
    {
      if (dil[it]>0 && dil[it]<=(*kmax))
	p->faces[ftrace->v].color=
	  mid+1+(int)((mid-2)*(dil[it]-1.0)/((*kmax)-1.0));
      else if (dil[it]>0)
	p->faces[ftrace->v].color=2; /* over ceiling? dark blue. */
      else
	p->faces[ftrace->v].color=1; /* couldn't compute? black */
      ftrace=ftrace->next;
      it++;
    }
  if (facelist) vert_free(&facelist);
  if (dil) free(dil);
  return count;
} /* color_qc_xyz */

int h_compare_area(struct p_data *p,struct p_data *q)
     /* compare ratio of hyperbolic areas of faces from p and q. 
In 2-color ramp, lower indices for q larger, uppers for p larger. */
{
  int i,node,mid;
  double b=1.0,ratio,*areas_p,*areas_q;
  struct R_data *pR_ptr,*qR_ptr;

  pR_ptr=p->packR_ptr;
  qR_ptr=q->packR_ptr;
  mid=(int)(color_ramp_size/2);
  node=(p->facecount>q->facecount) ? q->facecount : p->facecount;
  if (!( areas_p=(double *)malloc((node+1)*sizeof(double)) )
      || !( areas_q=(double *)malloc((node+1)*sizeof(double)) ) ) 
    return 0;
  for (i=1;i<=node;i++)
    {
      areas_p[i]=h_area(pR_ptr[p->faces[i].vert[0]].rad,
			pR_ptr[p->faces[i].vert[1]].rad,
			pR_ptr[p->faces[i].vert[2]].rad);
      areas_q[i]=h_area(qR_ptr[q->faces[i].vert[0]].rad,
			qR_ptr[q->faces[i].vert[1]].rad,
			qR_ptr[q->faces[i].vert[2]].rad);
      ratio=areas_q[i]/areas_p[i];
      if (ratio>b) b=ratio;
      if ((1.0/ratio)>b) b=1.0/ratio;
    }
  if (b==1) for (i=1;i<=p->facecount;i++) p->faces[i].color=mid;
  else
    {
      for (i=1;i<=node;i++)
	{
	  if ((ratio=areas_p[i]/areas_q[i])>1.0)
	    p->faces[i].color=
	      (int)(mid+(mid-1)*(ratio-1.0)/(b-1.0));
	  else 
	    p->faces[i].color=
	      1+(int)((mid-2)*(1.0-(1.0/ratio-1.0)/(b-1.0)));
	}
    }
  free(areas_p);free(areas_q);
  if (node<p->facecount) for (i=node+1;i<=p->facecount;i++)
    p->faces[i].color=mid;
  return 1;
} /* h_compare_area */

int h_compare_cl(struct p_data *p,struct p_data *q)
     /* compare ratio of hyp radii of packs p and q, put results into 
colors of pack p. Use 2-color ramp, lower indices for q larger, 
uppers for p larger. Data is scaled. */
{
  int i,node,mid;
  double b=1.0,ratio;
  struct K_data *pK_ptr;
  struct R_data *pR_ptr,*qR_ptr;

  pK_ptr=p->packK_ptr;pR_ptr=p->packR_ptr;
  qR_ptr=q->packR_ptr;
  mid=(int)(color_ramp_size/2);
  node=(p->nodecount>q->nodecount) ? q->nodecount : p->nodecount;
  for (i=1;i<=node;i++)
    {
      if (pR_ptr[i].rad>0.0 && qR_ptr[i].rad>0.0)
	{
	  ratio=log(pR_ptr[i].rad)/log(qR_ptr[i].rad);
	  if (ratio>b) b=ratio;
	  else if ((1.0/ratio)>b) b=1.0/ratio;
	}
    } /* determine b for scaling purposes */
  if (b==1.0) 
    {
      for (i=1;i<=node;i++)
	{
	  if (pR_ptr[i].rad<=0.0 && qR_ptr[i].rad>0.0)
	    pK_ptr[i].color=color_ramp_size;
	  else if (pR_ptr[i].rad>0.0 && qR_ptr[i].rad<=0.0)
	    pK_ptr[i].color=1;
	  else pK_ptr[i].color=mid;
	}
      return 1;
    }
  for (i=1;i<=node;i++)
    {
      if (pR_ptr[i].rad<=0.0 && qR_ptr[i].rad>0.0)
	pK_ptr[i].color=color_ramp_size;
      else if (pR_ptr[i].rad>0.0 && qR_ptr[i].rad<=0.0)
	pK_ptr[i].color=1;
      else
	{
	  ratio=log(pR_ptr[i].rad)/log(qR_ptr[i].rad);
	  if (ratio>1.0) pK_ptr[i].color=
			   (int)(mid+(mid-1)*(1.0-ratio)/(1.0-b));
	  else pK_ptr[i].color=
		 1+(int)((mid-2)*(1.0-(1.0-1.0/ratio)/(1.0-b)));
	}
    }
  if (node<p->nodecount) for (i=node+1;i<=p->nodecount;i++)
    pK_ptr[i].color=mid;
  return 1;
} /* h_compare_cl */

int e_compare_cl(struct p_data *p,struct p_data *q)
     /* compare ratio of eucl radii of packs p and q, put results into 
colors of pack p. Use 2-color ramp, lower indices indicating q larger, 
uppers indicating p larger. Data is scaled. */
{
  int i,node,mid;
  double b=1.0,ratio;
  struct K_data *pK_ptr;
  struct R_data *pR_ptr,*qR_ptr;

  pK_ptr=p->packK_ptr;pR_ptr=p->packR_ptr;
  qR_ptr=q->packR_ptr;

  mid=(int)(color_ramp_size/2);
  node=(p->nodecount>q->nodecount) ? q->nodecount : p->nodecount;
  for (i=1;i<=node;i++)
    {
      ratio=pR_ptr[i].rad/qR_ptr[i].rad;
      if (ratio>b) b=ratio;
      else if ((1/ratio)>b) b=1/ratio;
    } /* determine b for scaling purposes */
  if (b==1.0) 
    {
      for (i=1;i<=node;i++) pK_ptr[i].color=mid;
      return 1;
    }
  for (i=1;i<=node;i++)
    {
      ratio=pR_ptr[i].rad/qR_ptr[i].rad;
      if (ratio>1.0) pK_ptr[i].color=
		       (int)(mid+(mid-1)*(1.0-ratio)/(1.0-b));
      else pK_ptr[i].color=
	     1+(int)((mid-2)*(1.0-(1.0-1.0/ratio)/(1.0-b)));
    }
  if (node<p->nodecount) for (i=node+1;i<=p->nodecount;i++)
    pK_ptr[i].color=mid;
  return 1;
} /* e_compare_cl */

int e_compare_area(struct p_data *p,struct p_data *q)
     /* compare ratio of eucl areas of faces from p and q. Use 2-color 
ramp, lower indices indicating q larger, uppers indicating p larger. */
{
  int i,node,flag=1,mid;
  double b=1.0,ratio,*areas_p,*areas_q;
  struct R_data *pR_ptr,*qR_ptr;

  pR_ptr=p->packR_ptr;
  qR_ptr=q->packR_ptr;
  mid=(int)(color_ramp_size/2);
  node=(p->facecount>q->facecount) ? q->facecount : p->facecount;
  if (!(areas_p=(double *)malloc((node+1)*sizeof(double)) ) 
      || !(areas_q=(double *)malloc((node+1)*sizeof(double)) ) ) 
    return 0;
  for (i=1;i<=node;i++)
    {
      areas_p[i]=e_area(pR_ptr[p->faces[i].vert[0]].rad,
			pR_ptr[p->faces[i].vert[1]].rad,
			pR_ptr[p->faces[i].vert[2]].rad);
      areas_q[i]=e_area(qR_ptr[q->faces[i].vert[0]].rad,
			qR_ptr[q->faces[i].vert[1]].rad,
			qR_ptr[q->faces[i].vert[2]].rad);
      ratio=areas_q[i]/areas_p[i];
      if (ratio>b) b=ratio;
      if (1.0/ratio>b) b=1.0/ratio;
    }
  if (b==1) flag=0;
  else
    {
      for (i=1;i<=node;i++)
	{
	  if ((ratio=areas_p[i]/areas_q[i])>1.0)
	    p->faces[i].color=
	      (int)(mid+(mid-1)*(ratio-1.0)/(b-1.0));
	  else 
	    p->faces[i].color=
	      1+(int)((mid-2)*(1.0-(1.0/ratio-1.0)/(b-1.0)));
	}
    }
  free(areas_p);free(areas_q);
  if (node<p->facecount) for (i=node+1;i<=p->facecount;i++)
    p->faces[i].color=mid;
  return (flag);
} /* e_compare_area */
