#include "sa_types.h"
#include "sa_proto.h"

/* Main aim, to extract a 'skeleton' complex from parent complex p. 
Return new packing q or NULL on error.

  Skeleton is normally formed by circles along certain edges and
their immediate neighbors. 'cookie_cutter' doesn't work because it
prunes boundary vertices since they're not interior. Hence this routine. 

  A vertlist is passed containing the 'skeleton' vertices (but in fact any
set of vertices); we take these and all nghbs to form the new complex. 
We have to renumber, so we keep vertex_map pointing to the old numbers.

  Note, result may not be a legal complex: (1) vertices may have no
interior neighbors, e.g., bdry vertices. In applications, we still 
want to repack; usually bdry verts have target angle sums. (2) interior
may not be connected. (3) There may be boundary vertices of new complex
which have more than 2 boundary nghbs. (NOTE: this last may be the
worst problem, so I will search for it and abort.)

  Key aim is to get complex which can be repacked (even if it can't
be laid out).

*/

struct p_data *skeleton(struct p_data *p,struct Vertlist *vlist)
{
  int i,j,u,v,w,new_v,num,count=0,hits,*vert_flags;
  struct Vertlist *vtrace;
  struct Edgelist *etrace;
  struct p_data *q=NULL;
  struct K_data *pK_ptr=p->packK_ptr;
  struct R_data *pR_ptr=p->packR_ptr;
  struct K_data *qK_ptr=NULL;
  struct R_data *qR_ptr=NULL;

  if (!p->status || !vlist) {
    sprintf(msgbuf,"skeleton: packing or vertlist not given; abort.");
    emsg();
    return NULL;
  }

  /* mark relevant vertices */
  vert_flags=(int *)calloc((size_t)(p->nodecount+1),sizeof(int));
  vtrace=vlist;
  while(vtrace) {
    vert_flags[vtrace->v]=-1;
    vtrace=vtrace->next;
    count++;
  }
  if(!count) return NULL;
  for (v=1;v<=p->nodecount;v++) {
    if (vert_flags[v]<0)
      for (j=0;j<=pK_ptr[v].num;j++)
	if (vert_flags[(w=pK_ptr[v].flower[j])]>=0)
	  vert_flags[w]++;
  }
    
  /* if there is a vert with too many boundary nghbs, abort. */
  for (v=1;v<=p->nodecount;v++) 
    if (vert_flags[v]) {
      hits=0;
      if (pK_ptr[v].bdry_flag) {
	if (vert_flags[pK_ptr[v].flower[0]]) hits++;
	if (vert_flags[pK_ptr[v].flower[pK_ptr[v].num]]) hits++;
      }
      for (j=0;j<pK_ptr[v].num;j++) {
	u=pK_ptr[v].flower[j];
	w=pK_ptr[v].flower[j+1];
	if ((vert_flags[u] && !vert_flags[w]) 
	    || (!vert_flags[u] && vert_flags[w])) hits++;
      }
      if (hits>2) {
	sprintf(msgbuf,"skeleton: v%d has too many boundary "
		"neighbors; abort.",v);
	emsg();
	return NULL;
      }
    }

  /* set new vertex numbers */
  hits=0;
  for (i=1;i<=p->nodecount;i++)
    if (vert_flags[i]) vert_flags[i]=++hits;

  if (hits<4 || !(q=init_packing()) || !alloc_pack_space(q,hits+10,0)) {
    sprintf(msgbuf,"skeleton: Failed to form a packing: %d vertices.",hits);
    emsg();
    return NULL;
  }

  q->hes=p->hes;
  q->nodecount=hits;
  qK_ptr=q->packK_ptr;
  qR_ptr=q->packR_ptr;
  q->vertex_map=etrace=(struct Edgelist *)calloc(1,sizeof(struct Edgelist));
  
  /* form new flowers */
  for (v=1;v<=p->nodecount;v++)
    if ((new_v=vert_flags[v])>0) { /* this is one of the new vertices */
      hits=0;
      for (j=0;j<pK_ptr[v].num+pK_ptr[v].bdry_flag;j++)
	if (vert_flags[pK_ptr[v].flower[j]]) hits++;

      /* this should never happen */

      if (hits==1) {
	free_p_data(&q);
	sprintf(msgbuf,"skeleton: Combinatorial error.");
	emsg();
	return NULL;
      }

      qK_ptr[new_v].flower=(int *)calloc((size_t)(hits+1),sizeof(int));

      /* whole flower is in */
      if (hits==(pK_ptr[v].num+pK_ptr[v].bdry_flag)) { 
	for (j=0;j<=pK_ptr[v].num;j++)
	  qK_ptr[new_v].flower[j]=vert_flags[pK_ptr[v].flower[j]];
	qK_ptr[new_v].num=pK_ptr[v].num;
	/* qK_ptr[new_v].bdry_flag=pK_ptr[v].bdry_flag;*/
	qK_ptr[new_v].plot_flag=1;
      }

      else { /* partial flower; must find first new vert nghb. */
	num=pK_ptr[v].num+pK_ptr[v].bdry_flag;
	if (pK_ptr[v].bdry_flag) {
	  j=0;
	  while(!vert_flags[pK_ptr[v].flower[j]]) j++;
	}
	else {
	  j=0;
	  while(vert_flags[pK_ptr[v].flower[j]]) j++;
	  i=0;
	  while(!vert_flags[pK_ptr[v].flower[(i+j)%num]]) i++;
	  j=(i+j)%num;
	}
	for (i=0;i<hits;i++)
	  qK_ptr[new_v].flower[i]=vert_flags[pK_ptr[v].flower[(i+j)%num]];
	qK_ptr[new_v].num=hits-1;
	/* qK_ptr[new_v].bdry_flag=1;*/
	qK_ptr[new_v].plot_flag=1;
      }

      /* copy data */
      qR_ptr[new_v].rad=pR_ptr[v].rad;
      qR_ptr[new_v].aim=pR_ptr[v].aim;
      qR_ptr[new_v].center=pR_ptr[v].center;
      
      /* add to vertex_map */
      etrace=etrace->next=
	(struct Edgelist *)calloc(1,sizeof(struct Edgelist));
      etrace->v=new_v;
      etrace->w=v;
    }

  /* first vertex_map spot is empty */
  etrace=q->vertex_map;
  q->vertex_map=q->vertex_map->next;
  free(etrace);

  q->status=1;
  v=1;
  while (q->packK_ptr[v].flower[0]!=
	 q->packK_ptr[v].flower[q->packK_ptr[v].num]) v++;
  if (v>q->nodecount) {q->alpha=1;q->gamma=2;}
  else if (v==1) {q->alpha=1;q->gamma=2;}
  else {q->alpha=v;q->gamma=1;}
  return q;
} /* skeleton */

int skeleton_call(struct p_data *p,char *datastr)
{
  int hits;
  char *nextpoint,*endptr;
  struct Vertlist *vlist=NULL,*vtrace;
  struct p_data *q=NULL;

  nextpoint=datastr;
  stripsp(nextpoint);
  if (!p->status || !(vlist=Node_link_parse(p,datastr,&endptr,&hits,
      &Vlist,&Elist,&Flist,&region,pathlist,pathlength))) return 0;
  if ((q=skeleton(p,vlist)) && copy_pack(q,p)) return hits;
  if (q) free_p_data(&q);
  return 0;
} /* skeleton_call */



